
# ----------------------------------------------------------------------
#  COMPONENT: field - extracts data from an XML description of a field
#
#  This object represents one field in an XML description of a device.
#  It simplifies the process of extracting data vectors that represent
#  the field.
# ======================================================================
#  AUTHOR:  Michael McLennan, Purdue University
#  Copyright (c) 2004-2012  HUBzero Foundation, LLC
#
#  See the file "license.terms" for information on usage and
#  redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# ======================================================================
package require Itcl
package require BLT

namespace eval Rappture { # forward declaration }

itcl::class Rappture::Field {
    constructor {xmlobj path} { # defined below }
    destructor { # defined below }

    public method components {args}
    public method mesh {{cname -overall}}
    public method values {{cname -overall}}
    public method blob {{cname -overall}}
    public method limits {axis}
    public method controls {option args}
    public method hints {{key ""}}
    public method style { cname }
    public method isunirect2d {}
    public method isunirect3d {}
    public method extents {{cname -overall}}
    public method flowhints { cname }
    public method type {}
    public method viewer {} {
	return $_viewer 
    }
    public method vtkdata {cname}
    
    protected method _build {}
    protected method _getValue {expr}

    private variable _dim	0;	# Dimension of mesh
    private variable _xmlobj ""  ;      # ref to XML obj with device data
    private variable _path "";          # Path of this object in the XML 
    private variable _units ""   ;      # system of units for this field
    private variable _limits     ;# maps box name => {z0 z1} limits
    private variable _zmax 0     ;# length of the device

    private variable _field ""
    private variable _fieldNames ""   ;# list of field names.
    private variable _fieldUnits ""   ;# list of units of each field name.
    private variable _fieldLabels ""  ;# list of labels of each field name.
    private variable _comp2dims  ;# maps component name => dimensionality
    private variable _comp2xy    ;# maps component name => x,y vectors
    private variable _comp2vtk   ;# maps component name => vtkFloatArray
    private variable _comp2vtkstreamlines   ;# maps component name => vtkFloatArray
    private variable _comp2vtkcontour   ;# maps component name => vtkFloatArray
    private variable _comp2vtkvolume   ;# maps component name => vtkFloatArray
    private variable _comp2volume   ;# maps component name => vtkFloatArray
    private variable _comp2dx    ;# maps component name => OpenDX data
    private variable _comp2unirect2d ;# maps component name => unirect2d obj
    private variable _comp2unirect3d ;# maps component name => unirect3d obj
    private variable _comp2style ;# maps component name => style settings
    private variable _comp2cntls ;# maps component name => x,y control points
    private variable _comp2extents 
    private variable _comp2limits 
    private variable _type "" 
    private variable _comp2flowhints 
    private variable _comp2heightmap 
    private common _counter 0    ;# counter for unique vector names

    private method BuildPointsOnMesh { cname } 
    private method ConvertToVtkData { cname } 
    private method ReadVtkDataSet { cname contents } 
    private variable _isVtk
    private variable _values ""
    private variable _viewer "vtkcontour"
}

# ----------------------------------------------------------------------
# CONSTRUCTOR
# ----------------------------------------------------------------------
itcl::body Rappture::Field::constructor {xmlobj path} {
    if {![Rappture::library isvalid $xmlobj]} {
        error "bad value \"$xmlobj\": should be Rappture::library"
    }
    set _xmlobj $xmlobj
    set _path $path
    set _field [$xmlobj element -as object $path]
    set _units [$_field get units]

    set xunits [$xmlobj get units]
    if {"" == $xunits || "arbitrary" == $xunits} {
        set xunits "um"
    }

    # determine the overall size of the device
    set z0 [set z1 0]
    foreach elem [$_xmlobj children components] {
        switch -glob -- $elem {
            box* {
                if {![regexp {[0-9]$} $elem]} {
                    set elem "${elem}0"
                }
                set z0 [$_xmlobj get components.$elem.corner0]
                set z0 [Rappture::Units::convert $z0 \
                    -context $xunits -to $xunits -units off]

                set z1 [$_xmlobj get components.$elem.corner1]
                set z1 [Rappture::Units::convert $z1 \
                    -context $xunits -to $xunits -units off]

                set _limits($elem) [list $z0 $z1]
            }
        }
    }
    set _zmax $z1

    # build up vectors for various components of the field
    _build
}

# ----------------------------------------------------------------------
# DESTRUCTOR
# ----------------------------------------------------------------------
itcl::body Rappture::Field::destructor {} {
    itcl::delete object $_field
    # don't destroy the _xmlobj! we don't own it!

    foreach name [array names _comp2xy] {
        eval blt::vector destroy $_comp2xy($name)
    }
    foreach name [array names _comp2vtk] {
        set mobj [lindex $_comp2vtk($name) 0]
        set class [$mobj info class]
        ${class}::release $mobj

        set fobj [lindex $_comp2vtk($name) 1]
        rename $fobj ""
    }
    foreach name [array names _comp2unirect2d] {
        itcl::delete object $_comp2unirect2d($name)
    }
    foreach name [array names _comp2unirect3d] {
        itcl::delete object $_comp2unirect3d($name)
    }
    foreach name [array names _comp2flowhints] {
        itcl::delete object $_comp2flowhints($name)
    }
    foreach name [array names _comp2heightmap] {
        set mesh [lindex $_comp2heightmap($name) 0]
        set class [$mesh info class]
        ${class}::release $mesh

        set vtkarray [lindex $_comp2heightmap($name) 1]
        rename $vtkarray ""
    }
}

# ----------------------------------------------------------------------
# USAGE: components ?-name|-dimensions|-style? ?<pattern>?
#
# Returns a list of names or types for the various components of
# this field.  If the optional glob-style <pattern> is specified,
# then it returns only the components with names matching the pattern.
# ----------------------------------------------------------------------
itcl::body Rappture::Field::components {args} {
    Rappture::getopts args params {
        flag what -name default
        flag what -dimensions
        flag what -style
        flag what -particles
        flag what -flow
        flag what -box
    }

    set pattern *
    if {[llength $args] > 0} {
        set pattern [lindex $args 0]
        set args [lrange $args 1 end]
    }
    if {[llength $args] > 0} {
        error "wrong # args: should be \"components ?switches? ?pattern?\""
    }

    set rlist ""

    # BE CAREFUL: return component names in proper order
    foreach cname [$_field children -type component] {
        if {[info exists _comp2dims($cname)]
              && [string match $pattern $cname]} {

            switch -- $params(what) {
                -name { lappend rlist $cname }
                -dimensions { lappend rlist $_comp2dims($cname) }
                -style { lappend rlist $_comp2style($cname) }
            }
        }
    }
    return $rlist
}

# ----------------------------------------------------------------------
# USAGE: mesh ?<name>?
#
# Returns a list {xvec yvec} for the specified field component <name>.
# If the name is not specified, then it returns the vectors for the
# overall field (sum of all components).
# ----------------------------------------------------------------------
itcl::body Rappture::Field::mesh {{cname -overall}} {
    if {$cname == "-overall" || $cname == "component0"} {
        set cname [lindex [components -name] 0]
    }
    if {[info exists _comp2xy($cname)]} {
        return [lindex $_comp2xy($cname) 0]  ;# return xv
    }
    if { [info exists _comp2vtkstreamlines($cname)] } {
        error "mesh: not implemented for streamlines"
        return [$mobj mesh]
    }
    if { [info exists _comp2vtkcontour($cname)] } {
        error "method \"mesh\" is not implemented for vtkcontour"
    }
    if { [info exists _comp2vtk($cname)] } {
        set mobj [lindex $_comp2vtk($cname) 0]
        return [$mobj mesh]
    }
    if {[info exists _comp2dx($cname)]} {
        return ""  ;# no mesh -- it's embedded in the value data
    }
    if {[info exists _comp2heightmap($cname)]} {
        return ""  ;# no mesh -- it's embedded in the value data
    }
    if {[info exists _comp2vtkvolume($cname)]} {
        return ""  ;# no mesh -- it's embedded in the value data
    }
    if {[info exists _comp2unirect2d($cname)]} {
        set mobj [lindex $_comp2unirect2d($cname) 0]
        return [$mobj mesh]
    }
    if {[info exists _comp2unirect3d($cname)]} {
        set mobj [lindex $_comp2unirect3d($cname) 0]
        return [$mobj mesh]
    }
    error "bad option \"$cname\": should be [join [lsort [array names _comp2dims]] {, }]"
}

# ----------------------------------------------------------------------
# USAGE: values ?<name>?
#
# Returns a list {xvec yvec} for the specified field component <name>.
# If the name is not specified, then it returns the vectors for the
# overall field (sum of all components).
# ----------------------------------------------------------------------
itcl::body Rappture::Field::values {{cname -overall}} {
    if {$cname == "component0"} {
        set cname "component"
    }
    if {[info exists _comp2xy($cname)]} {
        return [lindex $_comp2xy($cname) 1]  ;# return yv
    }
    if { [info exists _comp2vtkcontour($cname)] } {
        error "method \"values\" is not implemented for vtkcontour"
    }
    if { [info exists _comp2heightmap($cname)] } {
        return $_values
    }
    if { [info exists _comp2vtkstreamlines($cname)] } {
        # FIXME: Need to process the vtk file data to pull out the field's
        # values.
        error "vtkstreamlines: values not implemented"
        return [lindex $_comp2vtkstreamlines($cname) 1] 
    }
    if { [info exists _comp2vtk($cname)] } {
        return [lindex $_comp2vtk($cname) 1]  ;# return vtkFloatArray
    }
    if {[info exists _comp2dx($cname)]} {
        return $_comp2dx($cname)  ;# return gzipped, base64-encoded DX data
    }
    if {[info exists _comp2unirect2d($cname)]} {
        return [$_comp2unirect2d($cname) values]
    }
    if {[info exists _comp2unirect3d($cname)]} {
        return [$_comp2unirect3d($cname) blob]
    }
    error "bad option \"$cname\": should be [join [lsort [array names _comp2dims]] {, }]"
}

# ----------------------------------------------------------------------
# USAGE: blob ?<name>?
#
# Returns a string representing the blob of data for the mesh and values.
# ----------------------------------------------------------------------
itcl::body Rappture::Field::blob {{cname -overall}} {
    if {$cname == "component0"} {
        set cname "component"
    }
    if {[info exists _comp2xy($cname)]} {
        return ""
    }
    if { [info exists _comp2vtk($cname)] } {
        return ""
    }
    if { [info exists _comp2vtkvolume($cname)] } {
        return $_comp2vtkvolume($cname)
    }
    if { [info exists _comp2vtkcontour($cname)] } {
        return $_comp2vtkcontour($cname)
    }
    if { [info exists _comp2vtkstreamlines($cname)] } {
        # Return the contents of the vtk file.
        return $_comp2vtkstreamlines($cname)
    }
    if {[info exists _comp2dx($cname)]} {
        return $_comp2dx($cname)  ;# return gzipped, base64-encoded DX data
    }
    if {[info exists _comp2unirect2d($cname)]} {
        return [$_comp2unirect2d($cname) blob]
    }
    if {[info exists _comp2unirect3d($cname)]} {
        return [$_comp2unirect3d($cname) blob]
    }
    error "bad option \"$cname\": should be [join [lsort [array names _comp2dims]] {, }]"
}

# ----------------------------------------------------------------------
# USAGE: limits <axis>
#
# Returns a list {min max} representing the limits for the specified
# axis.
# ----------------------------------------------------------------------
itcl::body Rappture::Field::limits {which} {
    set min ""
    set max ""

    blt::vector tmp zero
    foreach cname [array names _comp2dims] {
        switch -- $_comp2dims($cname) {
            1D {
                switch -- $which {
                    x - xlin { set pos 0; set log 0; set axis xaxis }
                    xlog { set pos 0; set log 1; set axis xaxis }
                    y - ylin - v - vlin { set pos 1; set log 0; set axis yaxis }
                    ylog - vlog { set pos 1; set log 1; set axis yaxis }
                    default {
                        error "bad option \"$which\": should be x, xlin, xlog, y, ylin, ylog, v, vlin, vlog"
                    }
                }

                set vname [lindex $_comp2xy($cname) $pos]
                $vname variable vec

                if {$log} {
                    # on a log scale, use abs value and ignore 0's
                    $vname dup tmp
                    $vname dup zero
                    zero expr {tmp == 0}            ;# find the 0's
                    tmp expr {abs(tmp)}             ;# get the abs value
                    tmp expr {tmp + zero*max(tmp)}  ;# replace 0's with abs max
                    set vmin [blt::vector expr min(tmp)]
                    set vmax [blt::vector expr max(tmp)]
                } else {
                    set vmin $vec(min)
                    set vmax $vec(max)
                }

                if {"" == $min} {
                    set min $vmin
                } elseif {$vmin < $min} {
                    set min $vmin
                }
                if {"" == $max} {
                    set max $vmax
                } elseif {$vmax > $max} {
                    set max $vmax
                }
            }
            2D - 3D {
                if {[info exists _comp2unirect2d($cname)]} {
                    set limits [$_comp2unirect2d($cname) limits $which]
                    foreach {vmin vmax} $limits break
                    set axis vaxis
                } elseif {[info exists _comp2unirect3d($cname)]} {
                    set limits [$_comp2unirect3d($cname) limits $which]
                    foreach {vmin vmax} $limits break
                    set axis vaxis
                } elseif {[info exists _comp2vtk($cname)]} {
                    foreach {xv yv} $_comp2vtk($cname) break
                    switch -- $which {
                        x - xlin - xlog {
                            foreach {vmin vmax} [$xv limits x] break
                            set axis xaxis
                        }
                        y - ylin - ylog {
                            foreach {vmin vmax} [$xv limits y] break
                            set axis yaxis
                        }
                        z - zlin - zlog {
                            foreach {vmin vmax} [$xv limits z] break
                            set axis zaxis
                        }
                        v - vlin - vlog {
                            catch {unset style}
                            array set style $_comp2style($cname)
                            if {[info exists style(-min)] && [info exists style(-max)]} {
                                # This component has its own hard-coded
                                # min/max range.  Ignore it for overall limits.
                                set vmin $min
                                set vmax $max
                            } else {
                                foreach {vmin vmax} [$yv GetRange] break
                            }
                            set axis vaxis
                        }
                        default {
                            error "bad option \"$which\": should be x, xlin, xlog, y, ylin, ylog, v, vlin, vlog"
                        }
                    }
                } elseif {[info exists _comp2heightmap($cname)]} {
		    foreach {mesh vector} $_comp2heightmap($cname) break
		    switch -- $which {
                        x - xlin - xlog {
                            set axis xaxis
                            foreach {vmin vmax} [$mesh limits x] break
                        }
                        y - ylin - ylog {
                            foreach {vmin vmax} [$mesh limits y] break
                            set axis yaxis
                        }
                        z - zlin - zlog {
                            foreach {vmin vmax} [$mesh limits z] break
                            set axis zaxis
                        }
                        v - vlin - vlog {
			    
			    set vmin [$vector min]
			    set vmax [$vector max]
			    set axis vaxis
			}
		    }
                } else {
                    set vmin 0  ;# HACK ALERT! must be OpenDX data
                    set vmax 1
                    set axis vaxis
                }
            }
        }
        if { "" == $min || $vmin < $min } {
            set min $vmin
        }
        if { "" == $max || $vmax > $max } {
            set max $vmax
        }
    }
    blt::vector destroy tmp zero
    set val [$_field get $axis.min]
    if {"" != $val && "" != $min} {
        if {$val > $min} {
            # tool specified this min -- don't go any lower
            set min $val
        }
    }

    set val [$_field get $axis.max]
    if {"" != $val && "" != $max} {
        if {$val < $max} {
            # tool specified this max -- don't go any higher
            set max $val
        }
    }
    return [list $min $max]
}

# ----------------------------------------------------------------------
# USAGE: controls get ?<name>?
# USAGE: controls validate <path> <value>
# USAGE: controls put <path> <value>
#
# Returns a list {path1 x1 y1 val1  path2 x2 y2 val2 ...} representing
# control points for the specified field component <name>.
# ----------------------------------------------------------------------
itcl::body Rappture::Field::controls {option args} {
    switch -- $option {
        get {
            set cname [lindex $args 0]
            if {[info exists _comp2cntls($cname)]} {
                return $_comp2cntls($cname)
            }
            return ""
        }
        validate {
            set path [lindex $args 0]
            set value [lindex $args 1]
            set units [$_xmlobj get $path.units]

            if {"" != $units} {
                set nv [Rappture::Units::convert \
                    $value -context $units -to $units -units off]
            } else {
                set nv $value
            }
            if {![string is double $nv]
                  || [regexp -nocase {^(inf|nan)$} $nv]} {
                error "Value out of range"
            }

            set rawmin [$_xmlobj get $path.min]
            if {"" != $rawmin} {
                set minv $rawmin
                if {"" != $units} {
                    set minv [Rappture::Units::convert \
                        $minv -context $units -to $units -units off]
                    set nv [Rappture::Units::convert \
                        $value -context $units -to $units -units off]
                }
                # fix for the case when the user tries to
                # compare values like minv=-500 nv=-0600
                set nv [format "%g" $nv]
                set minv [format "%g" $minv]

                if {$nv < $minv} {
                    error "Minimum value allowed here is $rawmin"
                }
            }

            set rawmax [$_xmlobj get $path.max]
            if {"" != $rawmax} {
                set maxv $rawmax
                if {"" != $units} {
                    set maxv [Rappture::Units::convert \
                        $maxv -context $units -to $units -units off]
                    set nv [Rappture::Units::convert \
                        $value -context $units -to $units -units off]
                }
                # fix for the case when the user tries to
                # compare values like maxv=-500 nv=-0600
                set nv [format "%g" $nv]
                set maxv [format "%g" $maxv]

                if {$nv > $maxv} {
                    error "Maximum value allowed here is $rawmax"
                }
            }

            return "ok"
        }
        put {
            set path [lindex $args 0]
            set value [lindex $args 1]
            $_xmlobj put $path.current $value
            _build
        }
        default {
            error "bad option \"$option\": should be get or put"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: hints ?<keyword>?
#
# Returns a list of key/value pairs for various hints about plotting
# this field.  If a particular <keyword> is specified, then it returns
# the hint for that <keyword>, if it exists.
# ----------------------------------------------------------------------
itcl::body Rappture::Field::hints {{keyword ""}} {
    set hints(fieldnames)  $_fieldNames
    set hints(fieldunits)  $_fieldUnits
    set hints(fieldlabels) $_fieldLabels
    foreach {key path} {
        camera          camera.position
        color           about.color
        default         about.default
        group           about.group
        label           about.label
        fieldnames      about.fieldnames
        fieldunits      about.fieldunits
        fieldlabels     about.fieldlabels
        scale           about.scale
        seeds           about.seeds
        style           about.style
        toolId          tool.id
        toolName        tool.name
        toolRevision    tool.version.application.revision
        type            about.type
        xlabel          about.xaxis.label
        ylabel          about.yaxis.label
        zlabel          about.zaxis.label
        xunits          about.xaxis.units
        yunits          about.yaxis.units
        zunits          about.zaxis.units
        units           units
        updir           updir
        vectors         about.vectors
    } {
        set str [$_field get $path]
        if {"" != $str} {
            set hints($key) $str
        }
    }
    # Set tool and path hints
    set hints(tool) [$_xmlobj get tool.name]
    set hints(path) $_path
    if 0 {
        # to be compatible with curve objects
        set hints(xlabel) "Position"
    }
    if {[info exists hints(group)] && [info exists hints(label)]} {
        # pop-up help for each curve
        set hints(tooltip) $hints(label)
    }

    if {$keyword != ""} {
        if {[info exists hints($keyword)]} {
            return $hints($keyword)
        }
        return ""
    }
    return [array get hints]
}

# ----------------------------------------------------------------------
# USAGE: _build
#
# Used internally to build up the vector representation for the
# field when the object is first constructed, or whenever the field
# data changes.  Discards any existing vectors and builds everything
# from scratch.
# ----------------------------------------------------------------------
itcl::body Rappture::Field::_build {} {

    # Discard any existing data
    foreach name [array names _comp2xy] {
        eval blt::vector destroy $_comp2xy($name)
    }
    foreach name [array names _comp2vtk] {
        set mobj [lindex $_comp2vtk($name) 0]
        set class [$mobj info class]
        ${class}::release $mobj

        set fobj [lindex $_comp2vtk($name) 1]
        rename $fobj ""
    }
    foreach name [array names _comp2unirect2d] {
        eval itcl::delete object $_comp2unirect2d($name)
    }
    foreach name [array names _comp2unirect3d] {
        eval itcl::delete object $_comp2unirect3d($name)
    }
    catch {unset _comp2xy}
    catch {unset _comp2vtk}
    catch {unset _comp2dx}
    catch {unset _comp2dims}
    catch {unset _comp2style}
    array unset _comp2volume
    array unset _comp2vtkstreamlines
    array unset _comp2vtkcontour
    array unset _comp2unirect2d
    array unset _comp2unirect3d
    array unset _comp2extents
    array unset _dataobj2type
    #
    # Scan through the components of the field and create
    # vectors for each part.
    #
    foreach cname [$_field children -type component] {
        set type ""
        if { ([$_field element $cname.constant] != "" &&
            [$_field element $cname.domain] != "") ||
            [$_field element $cname.xy] != ""} {
            set type "1D"
        } elseif {[$_field element $cname.mesh] != "" &&
            [$_field element $cname.values] != ""} {
            set type "points-on-mesh"
        } elseif {[$_field element $cname.vtk] != ""} {
	    set _isVtkData($cname) 1
	    set view [$_field get "about.view"]
	    switch -- $view {
		"streamlines" {
		    set type "vtkstreamlines"
		}
		"contour" - "heightmap" {
		    set type "vtkcontour"
		}
		default {
		    set type "vtk"
		}
            }
        } elseif {[$_field element $cname.opendx] != ""} {
            global env
            if { [info exists env(VTKVOLUME)] } {
                set type "vtkvolume"
            } else {
                set type "dx"
            }
        } elseif {[$_field element $cname.dx] != ""} {
            global env
            if { [info exists env(VTKVOLUME)] } {
                set type "vtkvolume"
            } else {
                set type "dx"
            }
        }
        set _comp2style($cname) ""
        
        # Save the extents of the component
        if { [$_field element $cname.extents] != "" } {
            set extents [$_field get $cname.extents]
        } else {
            set extents 1 
        }
        set _comp2extents($cname) $extents
        set _type $type
        if {$type == "1D"} {
            #
            # 1D data can be represented as 2 BLT vectors,
            # one for x and the other for y.
            #
            set xv ""
            set yv ""

            set val [$_field get $cname.constant]
            if {$val != ""} {
                set domain [$_field get $cname.domain]
                if {$domain == "" || ![info exists _limits($domain)]} {
                    set z0 0
                    set z1 $_zmax
                } else {
                    foreach {z0 z1} $_limits($domain) { break }
                }
                set xv [blt::vector create x$_counter]
                $xv append $z0 $z1

                foreach {val pcomp} [_getValue $val] break
                set yv [blt::vector create y$_counter]
                $yv append $val $val

                if {$pcomp != ""} {
                    set zm [expr {0.5*($z0+$z1)}]
                    set _comp2cntls($cname) \
                        [list $pcomp $zm $val "$val$_units"]
                }
            } else {
                set xydata [$_field get $cname.xy]
                if {"" != $xydata} {
                    set xv [blt::vector create x$_counter]
                    set yv [blt::vector create y$_counter]
                    set tmp [blt::vector create \#auto]
                    $tmp set $xydata
                    $tmp split $xv $yv
                    blt::vector destroy $tmp
                }
            }

            if {$xv != "" && $yv != ""} {
                # sort x-coords in increasing order
                $xv sort $yv

                set _comp2dims($cname) "1D"
                set _comp2xy($cname) [list $xv $yv]
                incr _counter
            }
        } elseif {$type == "points-on-mesh"} {
	    BuildPointsOnMesh $cname 
        } elseif {$type == "vtk"} {
            #
            # Extract native vtk data from the XML and use a reader
            # to load it.
            #
	    error "not implemented"
        } elseif {$type == "vtkstreamlines"} {
            set _comp2dims($cname) "3D"
            # Allow redirects to another element.
            set vtkdata [$_field get $cname.vtk]
	    set _comp2vtkstreamlines($cname) $vtkdata
            set _comp2style($cname) [$_field get $cname.style]
            incr _counter
        } elseif {$type == "vtkcontour"} {
            set _comp2dims($cname) "2D"
            # Allow redirects to another element.

            set data [$_field get $cname.vtk]
	    ReadVtkDataSet $cname $data
	    set _comp2vtkcontour($cname) $data
            set _comp2style($cname) [$_field get $cname.style]
            incr _counter
        } elseif {$type == "vtkvolume"} {
            set _comp2dims($cname) "3D"
            # Allow redirects to another element.
            set data [$_field get -decode no $cname.dx]
            set data [Rappture::encoding::decode -as zb64 $data]
            if 1 {
            set file "/tmp/$cname.dx"
            set f [open $file "w"]
            puts $f $data
            close $f
            }
            set data [Rappture::ConvertDxToVtk $data]
            if 1 {
            set file "/tmp/$cname.vtk"
            set f [open $file "w"]
            puts $f $data
            close $f
            }
            set _comp2vtkvolume($cname) $data
            set _comp2style($cname) [$_field get $cname.style]
            incr _counter
        } elseif {$type == "vtkstreamlines2"} {
            set _comp2dims($cname) "3D"
            set _comp2vtkstreamlines($cname) [$_field get $cname.vtk]
            set _comp2style($cname) [$_field get $cname.style]
            incr _counter
        } elseif {$type == "dx" } {
            #
            # HACK ALERT!  Extract gzipped, base64-encoded OpenDX
            # data.  Assume that it's 3D.  Pass it straight
            # off to the NanoVis visualizer.
            #
            set _comp2dims($cname) "3D"
            set _comp2dx($cname)  [$_field get -decode no $cname.dx]
            if 1 {
            set data  [$_field get -decode yes $cname.dx]
            set file "/tmp/junk.dx"
            set f [open $file "w"]
            puts $f $data
            close $f
            if { [string match "<ODX>*" $data] } {
                set data [string range $data 5 end]
                set _comp2dx($cname) \
                        [Rappture::encoding::encode -as zb64 $data]
            } 
            }
            set _comp2style($cname) [$_field get $cname.style]
            if {[$_field element $cname.flow] != ""} {
                set _comp2flowhints($cname) \
                    [Rappture::FlowHints ::\#auto $_field $cname $_units]
            }
            incr _counter
        } elseif {$type == "opendx"} {
            #
            # HACK ALERT!  Extract gzipped, base64-encoded OpenDX
            # data.  Assume that it's 3D.  Pass it straight
            # off to the NanoVis visualizer.
            #
            set _comp2dims($cname) "3D"
            set data [$_field get -decode yes $cname.opendx]
            set data "<ODX>$data"
            set data [Rappture::encoding::encode -as zb64 $data]
            set _comp2dx($cname) $data
            set _comp2style($cname) [$_field get $cname.style]
            if {[$_field element $cname.flow] != ""} {
                set _comp2flowhints($cname) \
                    [Rapture::FlowHints ::\#auto $_field $cname $_units]
            }
            incr _counter
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: _getValue <expr>
#
# Used internally to get the value for an expression <expr>.  Returns
# a list of the form {val parameterPath}, where val is the numeric
# value of the expression, and parameterPath is the XML path to the
# parameter representing the value, or "" if the <expr> does not
# depend on any parameters.
# ----------------------------------------------------------------------
itcl::body Rappture::Field::_getValue {expr} {
    #
    # First, look for the expression among the <parameter>'s
    # associated with the device.
    #
    set found 0
    foreach pcomp [$_xmlobj children parameters] {
        set id [$_xmlobj element -as id parameters.$pcomp]
        if {[string equal $id $expr]} {
            set val [$_xmlobj get parameters.$pcomp.current]
            if {"" == $val} {
                set val [$_xmlobj get parameters.$pcomp.default]
            }
            if {"" != $val} {
                set expr $val
                set found 1
                break
            }
        }
    }
    if {$found} {
        set pcomp "parameters.$pcomp"
    } else {
        set pcomp ""
    }

    if {$_units != ""} {
        set expr [Rappture::Units::convert $expr \
            -context $_units -to $_units -units off]
    }

    return [list $expr $pcomp]
}

#
# isunirect2d  --
#
# Returns if the field is a unirect2d object.  
#
itcl::body Rappture::Field::isunirect2d { } {
    return [expr [array size _comp2unirect2d] > 0]
}

#
# isunirect3d  --
#
# Returns if the field is a unirect3d object.  
#
itcl::body Rappture::Field::isunirect3d { } {
    return [expr [array size _comp2unirect3d] > 0]
}

#
# flowhints  --
#
# Returns the hints associated with a flow vector field.  
#
itcl::body Rappture::Field::flowhints { cname } {
    if { [info exists _comp2flowhints($cname)] } {
        return $_comp2flowhints($cname)
    }
    return ""
}

#
# style  --
#
# Returns the style associated with a component of the field.  
#
itcl::body Rappture::Field::style { cname } {
    if { [info exists _comp2style($cname)] } {
        return $_comp2style($cname)
    }
    return ""
}

#
# type  --
#
# Returns the style associated with a component of the field.  
#
itcl::body Rappture::Field::type {} {
    return $_type
}

#
# extents --
#
# Returns if the field is a unirect2d object.  
#
itcl::body Rappture::Field::extents {{cname -overall}} {
    if {$cname == "-overall" } {
        set max 0
        foreach cname [$_field children -type component] {
            if { ![info exists _comp2unirect3d($cname)] &&
                 ![info exists _comp2extents($cname)] } {
                continue
            }
            set value $_comp2extents($cname)
            if { $max < $value } {
                set max $value
            }
        }
        return $max
    } 
    if { $cname == "component0"} {
        set cname [lindex [components -name] 0]
    }
    return $_comp2extents($cname)
}

itcl::body Rappture::Field::ConvertToVtkData { cname } {
    set ds ""
    switch -- [typeof $cname] {
	"unirect2d" {
	    foreach { x1 x2 xN y1 y2 yN } [$dataobj mesh $cname] break
	    set spacingX [expr {double($x2 - $x1)/double($xN - 1)}]
	    set spacingY [expr {double($y2 - $y1)/double($yN - 1)}]
	    
	    set ds [vtkImageData $this-grdataTemp]
	    $ds SetDimensions $xN $yN 1
	    $ds SetOrigin $x1 $y1 0
	    $ds SetSpacing $spacingX $spacingY 0
	    set arr [vtkDoubleArray $this-arrTemp]
	    foreach {val} [$dataobj values $cname] {
		$arr InsertNextValue $val
	    }
	    [$ds GetPointData] SetScalars $arr
	}
	"unirect3d" {
	    foreach { x1 x2 xN y1 y2 yN z1 z2 zN } [$dataobj mesh $cname] break
	    set spacingX [expr {double($x2 - $x1)/double($xN - 1)}]
	    set spacingY [expr {double($y2 - $y1)/double($yN - 1)}]
	    set spacingZ [expr {double($z2 - $z1)/double($zN - 1)}]
	    
	    set ds [vtkImageData $this-grdataTemp]
	    $ds SetDimensions $xN $yN $zN
	    $ds SetOrigin $x1 $y1 $z1
	    $ds SetSpacing $spacingX $spacingY $spacingZ
	    set arr [vtkDoubleArray $this-arrTemp]
	    foreach {val} [$dataobj values $cname] {
		$arr InsertNextValue $val
	    }
	    [$ds GetPointData] SetScalars $val
	}
	"vtkcontour" {
	    return [$dataobj blob $cname]
	}
	"dx" {
            return [Rappture::ConvertDxToVtk $_comp2dx($cname)]
	}
	default {
	    set mesh [$dataobj mesh $cname]
	    switch -- [$mesh GetClassName] {
		vtkPoints {
		    # handle cloud of points
		    set ds [vtkPolyData $this-polydataTemp]
		    $ds SetPoints $mesh
		    [$ds GetPointData] SetScalars [$dataobj values $cname]
		}
		vtkPolyData {
		    set ds [vtkPolyData $this-polydataTemp]
		    $ds ShallowCopy $mesh
		    [$ds GetPointData] SetScalars [$dataobj values $cname]
		}
		vtkUnstructuredGrid {
		    # handle 3D grid with connectivity
		    set ds [vtkUnstructuredGrid $this-grdataTemp]
		    $ds ShallowCopy $mesh
		    [$ds GetPointData] SetScalars [$dataobj values $cname]
		}
		vtkRectilinearGrid {
		    # handle 3D grid with connectivity
		    set ds [vtkRectilinearGrid $this-grdataTemp]
		    $ds ShallowCopy $mesh
		    [$ds GetPointData] SetScalars [$dataobj values $cname]
		}
		default {
		    error "don't know how to handle [$mesh GetClassName] data"
		}
	    }
	}
    }

    if {"" != $ds} {
        set writer [vtkDataSetWriter $this-dsWriterTmp]
        $writer SetInput $ds
        $writer SetFileTypeToASCII
        $writer WriteToOutputStringOn
        $writer Write
        set out [$writer GetOutputString]
        $ds Delete
        $writer Delete
    } else {
        set out ""
        error "No DataSet to write"
    }

    append out "\n"
    return $out
}

itcl::body Rappture::Field::ReadVtkDataSet { cname contents } {
    package require vtk

    set reader $this-datasetreader
    vtkDataSetReader $reader

    # Write the contents to a file just in case it's binary.
    set tmpfile file[pid].vtk
    set f [open "$tmpfile" "w"]
    fconfigure $f -translation binary -encoding binary
    puts $f $contents 
    close $f
    $reader SetFileName $tmpfile
    $reader ReadAllScalarsOn
    $reader ReadAllVectorsOn
    $reader ReadAllFieldsOn
    $reader Update
    set dataset [$reader GetOutput]
    set limits {}
    foreach {xmin xmax ymin ymax zmin zmax} [$dataset GetBounds] break
    # Figure out the dimension of the mesh from the bounds.
    set _dim 0
    if { $xmax > $xmin } {
	incr _dim
    }
    if { $ymax > $ymin } {
	incr _dim
    }
    if { $zmax > $zmin } {
	incr _dim
    }
    lappend limits xmin $xmin xmax $xmax ymin $ymin ymax $ymax  
    set dataAttrs [$dataset GetPointData]
    if { $dataAttrs == ""} {
	puts stderr "No point data"
    }
    for {set i 0} {$i < [$dataAttrs GetNumberOfArrays] } {incr i} {
	set array [$dataAttrs GetArray $i]
	set name  [$dataAttrs GetArrayName $i]
	foreach {min max} [$array GetRange] break
	lappend limits $name-min $min $name-max $max
	lappend _fieldNames $name 
	lappend _fieldUnits ""
	lappend _fieldLabels $name
    }
    set _comp2limits($cname) $limits
    $reader Delete
    file delete $tmpfile
}

#
# vtkdata ?<cname>?
#
#	Returns a string representing the mesh and field data for a specific
#	component in basic VTK file format.
#
itcl::body Rappture::Field::vtkdata {cname} {
    if {$cname == "component0"} {
        set cname "component"
    }
    if { [info exists _comp2vtkstreamlines($cname)] } {
        return $_comp2vtkstreamlines($cname)
    }
    if { [info exists _comp2vtkvolume($cname)] } {
        return $_comp2vtkvolume($cname)
    }
    if {[info exists _comp2dx($cname)]} {
        return $_comp2dx($cname)
    }
    if {[info exists _comp2unirect3d($cname)]} {
        return [$_comp2unirect3d($cname) blob]
    }
    if { [info exists _comp2vtkcontour($cname)] && $_dim == 2 } {
        return $_comp2vtkcontour($cname)
    }
    if { [info exists _comp2heightmap($cname)] } {
	set label [hints zlabel] 
	if { $label == "" } {
	    set label $cname
	} else {
	    regsub -all { } $label {_} label
	}
	foreach {mesh vector} $_comp2heightmap($cname) break
	append out "# vtk DataFile Version 3.0\n"
	append out "[hints label]\n"
	append out "ASCII\n"
	append out [$mesh vtkdata]
	append out "SCALARS $label float\n"
	append out "LOOKUP_TABLE default\n"
	append out "[$vector range 0 end]\n"
	return $out
    }
    error "this method should only be called by the vtkheightmap widget"
}

itcl::body Rappture::Field::BuildPointsOnMesh {cname} {
    #
    # More complex 2D/3D data is represented by a mesh
    # object and an associated vtkFloatArray for field
    # values.
    #
    set path [$_field get $cname.mesh]
    if {[$_xmlobj element $path] == ""} {
	# Unknown mesh designated.
	return
    }
    set element [$_xmlobj element -as type $path]
    lappend _fieldNames $cname
    lappend _fieldLabels $cname
    lappend _fieldUnits ""
    # Handle bizarre cases that hopefully will be deprecated.
    if { $element == "unirect3d" } {
        if { [$_field element $cname.extents] != "" } {
            set extents [$_field get $cname.extents]
        } else {
            set extents 1 
        }
	set _comp2dims($cname) "3D"
	set _comp2unirect3d($cname) \
	    [Rappture::Unirect3d \#auto $_xmlobj $_field $cname $extents]
	set _comp2style($cname) [$_field get $cname.style]
	if {[$_field element $cname.flow] != ""} {
	    set _comp2flowhints($cname) \
		[Rappture::FlowHints ::\#auto $_field $cname $_units]
	}
	incr _counter
	return
    }
    if { $element == "unirect2d" && [$_field element $cname.flow] != "" } {
        if { [$_field element $cname.extents] != "" } {
            set extents [$_field get $cname.extents]
        } else {
            set extents 1 
        }
	set _comp2dims($cname) "2D"
	set _comp2unirect2d($cname) \
	    [Rappture::Unirect2d \#auto $_xmlobj $path]
	set _comp2style($cname) [$_field get $cname.style]
	set _comp2flowhints($cname) \
	    [Rappture::FlowHints ::\#auto $_field $cname $_units]
	set _values [$_field element $cname.values]
	incr _counter
	return
    }
    switch -- $element {
	"cloud" {
	    set mesh [Rappture::Cloud::fetch $_xmlobj $path]
	}
	"mesh" {
	    set mesh [Rappture::Mesh::fetch $_xmlobj $path]
	}	    
	"unirect2d" {
	    set mesh [Rappture::Unirect2d::fetch $_xmlobj $path]
	    set _viewer "vtkheightmap"
	}
    }
    if {[$mesh dimensions] == 1} {
	# Is this used anywhere?
	#
	# OOPS!  This is 1D data
	# Forget the cloud/field -- store BLT vectors
	#
	set xv [blt::vector create x$_counter]
	set yv [blt::vector create y$_counter]
	
	set points [$mesh points]
	set max [$points GetNumberOfPoints]
	for {set i 0} {$i < $max} {incr i} {
	    $xv append [lindex [$points GetPoint $i] 0]
	}
	set class [$mesh info class]
	${class}::release $mesh
	
	set values [$_field get $cname.values]
	foreach y $values {
	    if {"" != $_units} {
		set y [Rappture::Units::convert $y \
			   -context $_units -to $_units -units off]
	    }
	    $yv append $y
	}
	# sort x-coords in increasing order
	$xv sort $yv
	
	set _comp2dims($cname) "1D"
	set _comp2xy($cname) [list $xv $yv]
	incr _counter
	return
    }

    if {[$mesh dimensions] == 3} {
	#
	# 3D data: Store cloud/field as components
	#
	set values [$_field get $cname.values]
	set farray [vtkFloatArray ::vals$_counter]
	foreach v $values {
	    if {"" != $_units} {
		set v [Rappture::Units::convert $v \
			   -context $_units -to $_units -units off]
	    }
	    $farray InsertNextValue $v
	}
	set _comp2dims($cname) "[$mesh dimensions]D"
	set _comp2vtk($cname) [list $mobj $farray]
	set _comp2style($cname) [$_field get $cname.style]
	incr _counter
	return
    }
    if {[$mesh dimensions] == 2} {

	# 2D data: Convert field data to vtkFloatArray

	set vector [blt::vector create \#auto]
	$vector set [$_field get $cname.values]
	set _comp2dims($cname) "[$mesh dimensions]D"
	set _comp2heightmap($cname) [list $mesh $vector]
	set _comp2style($cname) [$_field get $cname.style]
	incr _counter
	set label [hints zlabel]
	if { $label != "" } {
	    set _fieldNames $label
	    regsub -all { } $_fieldNames {_} _fieldNames
	    set _fieldLabels $label
	}
	set units [hints zunits]
	if { $units != "" } {
	    set _fieldUnits $units
	}
	return
    }
    error "unhandled case in field"
}
