
# ----------------------------------------------------------------------
#  COMPONENT: mesh - represents a structured mesh for a device
#
#  This object represents a mesh in an XML description of simulator
#  output.  A mesh is a structured arrangement of points, as elements
#  composed of nodes representing coordinates.  This is a little
#  different from a cloud, which is an unstructured arrangement
#  (shotgun blast) of points.
# ======================================================================
#  AUTHOR:  Michael McLennan, Purdue University
#  Copyright (c) 2004-2012  HUBzero Foundation, LLC
#
#  See the file "license.terms" for information on usage and
#  redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# ======================================================================
package require Itcl

namespace eval Rappture { 
    # forward declaration 
}

itcl::class Rappture::Mesh {
    private variable _xmlobj ""  ;	# Ref to XML obj with device data
    private variable _mesh ""    ;	# Lib obj representing this mesh
    private variable _dim	0;	# Dimension of mesh (1, 2, or 3)
    private variable _type "";		# Indicates the type of mesh.
    private variable _units "m m m" ;	# System of units for x, y, z
    private variable _limits        ;	# Array of mesh limits. Keys are 
					# xmin, xmax, ymin, ymax, ...
    private variable _numPoints 0   ;	# # of points in mesh
    private variable _vtkdata "";	# Mesh in vtk file format.
    constructor {xmlobj path} { 
	# defined below 
    }
    destructor { 
	# defined below 
    }
    public method points {}
    public method elements {}
    public method mesh {{-type "vtk"}}
    public method size {{what -points}}
    public method dimensions {}
    public method limits {which}
    public method hints {{key ""}}

    public proc fetch {xmlobj path}
    public proc release {obj}
    public method vtkdata {}
    public method type {} {
	return $_type
    }
    public method numpoints {} {
	return $_numPoints
    }

    private method ReadNodesElements {xmlobj path}
    private method _buildRectMesh {xmlobj path}
    private method _getVtkElement {npts}

    private variable _pts2elem   ;	# maps # points => vtk element

    private variable _points ""      ;	# name of vtkPointData object

    private common _xp2obj       ;	# used for fetch/release ref counting
    private common _obj2ref      ;	# used for fetch/release ref counting

    private variable _vtkoutput	""
    private variable _vtkreader	""
    private variable _vtkpoints	""
    private variable _vtkgrid	""
    private variable _xv	""
    private variable _yv	""
    private variable _zv	""
    private variable _bltmesh	""
    private variable _xValues	"";	# For the blt contour only
    private variable _yValues	"";	# For the blt contour only
    
    private method ReadCloud { xmlobj path }
    private method ReadGrid { xmlobj path }
    private method ReadTriangles { xmlobj path }
    private method ReadVtk { xmlobj path }
    private method GetBltCloudMesh { path }
    private method GetBltGridMesh { path }
    private method GetBltTriangleMesh { path }
    private method GetBltMesh { path } 
    private method GetDouble { path } 
    private method GetInt { path } 
    private method GetDimension {} 
    private method _buildNodesElements {xmlobj path} 
}

# ----------------------------------------------------------------------
# USAGE: Rappture::Mesh::fetch <xmlobj> <path>
#
# Clients use this instead of a constructor to fetch the Mesh for
# a particular <path> in the <xmlobj>.  When the client is done with
# the mesh, he calls "release" to decrement the reference count.
# When the mesh is no longer needed, it is cleaned up automatically.
# ----------------------------------------------------------------------
itcl::body Rappture::Mesh::fetch {xmlobj path} {
    set handle "$xmlobj|$path"
    if {[info exists _xp2obj($handle)]} {
        set obj $_xp2obj($handle)
        incr _obj2ref($obj)
        return $obj
    }
    set obj [Rappture::Mesh ::#auto $xmlobj $path]
    set _xp2obj($handle) $obj
    set _obj2ref($obj) 1
    return $obj
}

# ----------------------------------------------------------------------
# USAGE: Rappture::Mesh::release <obj>
#
# Clients call this when they're no longer using a Mesh fetched
# previously by the "fetch" proc.  This decrements the reference
# count for the mesh and destroys the object when it is no longer
# in use.
# ----------------------------------------------------------------------
itcl::body Rappture::Mesh::release {obj} {
    if {[info exists _obj2ref($obj)]} {
        incr _obj2ref($obj) -1
        if {$_obj2ref($obj) <= 0} {
            unset _obj2ref($obj)
            foreach handle [array names _xp2obj] {
                if {$_xp2obj($handle) == $obj} {
                    unset _xp2obj($handle)
                }
            }
            itcl::delete object $obj
        }
    } else {
        error "can't find reference count for $obj"
    }
}

# ----------------------------------------------------------------------
# CONSTRUCTOR
# ----------------------------------------------------------------------
itcl::body Rappture::Mesh::constructor {xmlobj path} {
    package require vtk
    if {![Rappture::library isvalid $xmlobj]} {
        error "bad value \"$xmlobj\": should be Rappture::library"
    }
    set _xmlobj $xmlobj
    set _mesh [$xmlobj element -as object $path]

    # Initialize mesh bounds to empty
    foreach lim {xmin xmax ymin ymax zmin zmax} {
        set _limits($lim) ""
    }
    GetDimension
    set u [$_mesh get units]
    if {"" != $u} {
        while {[llength $u] < 3} {
            lappend u [lindex $u end]
        }
        set _units $u
    }

    foreach lim {xmin xmax ymin ymax zmin zmax} {
        set _limits($lim) ""
    }
    # Meshes comes in a variety of flavors
    #
    # Dimensionality is determined from the <dimension> tag.  
    #
    # <vtk> described mesh
    # <element> +  <node> definitions
    # <cloud>		x,y coordinates or x,y,z coordinates 
    # <unirect>		uniform rectangular grid (2D) or cubes (3D)
    # <rectangle>	non-uniform rectangular mesh 
    if { [$_mesh element vtk] != ""} {
	ReadVtk $xmlobj $path
    } elseif { [$_mesh element cloud] != "" } {
	ReadCloud $xmlobj $path
    } elseif {[$_mesh element grid] != "" } {
	ReadGrid $xmlobj $path 
    } elseif {[$_mesh element triangles] != "" } {
	ReadTriangles $xmlobj $path
    } elseif {[$_mesh element node] != "" && [$_mesh element element] != ""} {
        ReadNodesElements $xmlobj $path
    } else {
        error "can't find mesh data in $path"
    }
}

# ----------------------------------------------------------------------
# DESTRUCTOR
# ----------------------------------------------------------------------
itcl::body Rappture::Mesh::destructor {} {
    # don't destroy the _xmlobj! we don't own it!
    itcl::delete object $_mesh

    if { $_xValues != "" } {
	blt::vector destroy $_xValues
    }
    if { $_yValues != "" } {
	blt::vector destroy $_yValues
    }
    if { $_vtkoutput != "" } {
	rename $_vtkoutput ""
    }
    if { $_vtkreader != "" } {
	rename $_vtkreader ""
    }
    if { $_points != "" } {
	rename $_points ""
    }
    if { $_bltmesh != "" } {
	blt::mesh destroy $_bltmesh
    }
    catch {rename $this-points ""}
    catch {rename $this-grid ""}
    catch {rename $this-dset ""}
    foreach type [array names _pts2elem] {
        rename $_pts2elem($type) ""
    }
}

#
# vtkdata -- 
#
#	This is called by the field object to generate a VTK file to send to
#	the remote render server.  Returns the vtkDataSet object containing
#	(at this point) just the mesh.  The field object doesn't know (or
#	care) what type of mesh is used.  The field object will add field
#	arrays before generating output to send to the remote render server.
#
itcl::body Rappture::Mesh::vtkdata {} {
    return $_vtkdata
}

# ----------------------------------------------------------------------
# USAGE: points
#
# Returns the vtk object containing the points for this mesh.
# ----------------------------------------------------------------------
itcl::body Rappture::Mesh::points {} {
    return $_points
}

# ----------------------------------------------------------------------
# USAGE: elements
#
# Returns a list of the form {plist r plist r ...} for each element
# in this mesh.  Each plist is a list of {x y x y ...} coordinates
# for the mesh.
# ----------------------------------------------------------------------
itcl::body Rappture::Mesh::elements {} {
    # build a map for region number => region type
    foreach comp [$_mesh children -type region] {
        set id [$_mesh element -as id $comp]
        set regions($id) [$_mesh get $comp]
    }
    set regions() "unknown"

    set rlist ""
    foreach comp [$_mesh children -type element] {
        set rid [$_mesh get $comp.region]

        #
        # HACK ALERT!
        #
        # Prophet puts out nodes in a funny "Z" shaped order,
        # not in proper clockwise fashion.  Switch the last
        # two nodes for now to make them correct.
        #
        set nlist [$_mesh get $comp.nodes]
        set n2 [lindex $nlist 2]
        set n3 [lindex $nlist 3]
        set nlist [lreplace $nlist 2 3 $n3 $n2]
        lappend nlist [lindex $nlist 0]

        set plist ""
        foreach nid $nlist {
            eval lappend plist [$_mesh get node($nid)]
        }
        lappend rlist $plist $regions($rid)
    }
    return $rlist
}

# ----------------------------------------------------------------------
# USAGE: mesh
#
# Returns the vtk object representing the mesh.
# ----------------------------------------------------------------------
itcl::body Rappture::Mesh::mesh { {type "vtk"} } {
    switch $type {
	"vtk" { 
	    return $_vtkgrid 
	}
	"blt" { 
	    if { $_bltmesh == "" } {
		GetBltMesh
		return $_bltmesh
	    }
	}
	default { 
	    error "Requested mesh type \"$type\" is unknown."
	}
    }
}

# ----------------------------------------------------------------------
# USAGE: size ?-points|-elements?
#
# Returns the number of points in this mesh.
# ----------------------------------------------------------------------
itcl::body Rappture::Mesh::size {{what -points}} {
    switch -- $what {
        -points {
            return [$_vtkpoints GetNumberOfPoints]
        }
        -elements {
            return [$_vtkgrid GetNumberOfCells]
        }
        default {
            error "bad option \"$what\": should be -points or -elements"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: dimensions
#
# Returns the number of dimensions for this object: 1, 2, or 3.
# ----------------------------------------------------------------------
itcl::body Rappture::Mesh::dimensions {} {
    return $_dim
}

# ----------------------------------------------------------------------
# USAGE: limits x|y|z
#
# Returns the {min max} values for the limits of the specified axis.
# ----------------------------------------------------------------------
itcl::body Rappture::Mesh::limits {which} {
    if {![info exists _limits(${which}min)]} {
        error "bad axis \"$which\": should be x, y, z"
    }
    return [list $_limits(${which}min) $_limits(${which}max)]
}

# ----------------------------------------------------------------------
# USAGE: hints ?<keyword>?
#
# Returns a list of key/value pairs for various hints about plotting
# this field.  If a particular <keyword> is specified, then it returns
# the hint for that <keyword>, if it exists.
# ----------------------------------------------------------------------
itcl::body Rappture::Mesh::hints {{keyword ""}} {
    foreach key {label color units} {
        set str [$_mesh get $key]
        if {"" != $str} {
            set hints($key) $str
        }
    }

    if {$keyword != ""} {
        if {[info exists hints($keyword)]} {
            return $hints($keyword)
        }
        return ""
    }
    return [array get hints]
}

itcl::body Rappture::Mesh::GetDimension {} {
    set string [$_xmlobj get dimension]
    if { [scan $string "%d" _dim] != 1 } {
	set _dim -1
	return
    }
    if { $_dim < 1 || $_dim > 4 } {
	error "bad dimension in mesh"
    }  
}

itcl::body Rappture::Mesh::GetDouble { path } {
    set string [$_xmlobj get $path]
    if { [scan $string "%g" value] != 1 } {
        puts stderr "can't get double value \"$string\" of \"$path\""
        return 0.0
    }
    return $value
}

itcl::body Rappture::Mesh::GetInt { path } {
    set string [$_xmlobj get $path]
    if { [scan $string "%d" value] != 1 } {
        puts stderr "can't get integer value \"$string\" of \"$path\""
        return 0.0
    }
    return $value
}

# ----------------------------------------------------------------------
# USAGE: _buildNodesElements <xmlobj> <path>
#
# Used internally to build a mesh representation based on nodes and
# elements stored in the XML.
# ----------------------------------------------------------------------
itcl::body Rappture::Mesh::_buildNodesElements {xmlobj path} {
    # create the vtk objects containing points and connectivity

    set _vtkpoints $this-points
    vtkPoints $_vtkpoints

    #
    # Extract each node and add it to the points list.
    #
    foreach comp [$xmlobj children -type node $path] {
        set xyz [$xmlobj get $path.$comp]

        # make sure we have x,y,z
        while {[llength $xyz] < 3} {
            lappend xyz "0"
        }

        # extract each point and add it to the points list
        foreach {x y z} $xyz break
        foreach dim {x y z} units $_units {
            set v [Rappture::Units::convert [set $dim] \
                -context $units -to $units -units off]

            set $dim $v  ;# save back to real x/y/z variable

            if {"" == $_limits(${dim}min)} {
                set _limits(${dim}min) $v
                set _limits(${dim}max) $v
            } else {
                if {$v < $_limits(${dim}min)} { set _limits(${dim}min) $v }
                if {$v > $_limits(${dim}max)} { set _limits(${dim}max) $v }
            }
        }
        $_vtkpoints InsertNextPoint $x $y $z
    }

    set _vtkgrid $this-grid
    vtkUnstructuredGrid $_vtkgrid
    $_vtkgrid SetPoints $_vtkpoints

    #
    # Extract each element and add it to the mesh.
    #
    foreach comp [$xmlobj children -type element $path] {
        set nlist [$_mesh get $comp.nodes]
        set elem [_getVtkElement [llength $nlist]]

        set i 0
        foreach n $nlist {
            [$elem GetPointIds] SetId $i $n
            incr i
        }
        $_vtkgrid InsertNextCell [$elem GetCellType] [$elem GetPointIds]
    }
}

# ----------------------------------------------------------------------
# USAGE: _buildRectMesh <xmlobj> <path>
#
# Used internally to build a mesh representation based on a native
# VTK file for a rectangular mesh.
# ----------------------------------------------------------------------
itcl::body Rappture::Mesh::_buildRectMesh {xmlobj path} {
    vtkRectilinearGridReader $this-gr
    $this-gr ReadFromInputStringOn
    $this-gr SetInputString [$xmlobj get $path.vtk]

    set _vtkgrid [$this-gr GetOutput]
    set _points [$_vtkgrid GetPointData]

    $_vtkgrid Update
    foreach name {xmin xmax ymin ymax zmin zmax} val [$_vtkgrid GetBounds] {
        set _limits($name) $val
    }
}

# ----------------------------------------------------------------------
# USAGE: _getVtkElement <npts>
#
# Used internally to find (or allocate, if necessary) a vtk element
# that can be used to build up a mesh.  The element depends on the
# number of points passed in.  4 points is a tetrahedron, 5 points
# is a pyramid, etc.
# ----------------------------------------------------------------------
itcl::body Rappture::Mesh::_getVtkElement {npts} {
    if {![info exists _pts2elem($npts)]} {
        switch -- $npts {
            3 {
                set _pts2elem($npts) $this-elem3
                vtkTriangle $_pts2elem($npts)
            }
            4 {
                set _pts2elem($npts) $this-elem4
                vtkTetra $_pts2elem($npts)
            }
            5 {
                set _pts2elem($npts) $this-elem5
                vtkPyramid $_pts2elem($npts)
            }
            6 {
                set _pts2elem($npts) $this-elem6
                vtkWedge $_pts2elem($npts)
            }
            8 {
                set _pts2elem($npts) $this-elem8
                vtkVoxel $_pts2elem($npts)
            }
        }
    }
    return $_pts2elem($npts)
}

itcl::body Rappture::Mesh::ReadCloud { xmlobj path } {
    set _type "cloud"
    set data [$xmlobj get $path.cloud.points] 
    Rappture::ReadPoints $data _dim values
	
    set all [blt::vector create \#auto]
    $all set $values 
    set xv [blt::vector create \#auto]
    set yv [blt::vector create \#auto]
    set zv [blt::vector create \#auto]
    if { $_dim == 2 } {
	$all split $xv $yv
	$zv seq 0 0 [$xv length]
    } elseif { $_dim == 3 } {
	$all split $xv $yv $zv
    } else {
	error "invalid dimension \"$_dim\" of cloud mesh"
    }
    set _numPoints [$xv length]
    if { $_dim == 2 } {
	append out "DATASET POLYDATA\n"
	append out "POINTS $_numPoints float\n"
	foreach x [$xv range 0 end] y [$yv range 0 end]  {
	    append out "$x $y 0\n"
	}
	append out "POINT_DATA $_numPoints\n"
	set _vtkdata $out
	foreach axis {x y} {
	    set vector [set ${axis}v]
	    set _limits(${axis}min) [${vector} min]
	    set _limits(${axis}max) [${vector} max]
	}
    } elseif { $_dim == 3 } {
	append out "DATASET POLYDATA\n"
	append out "POINTS $_numPoints float\n"
	foreach x [$xv range 0 end] y [$yv range 0 end] z [$zv range 0 end] {
	    append out "$x $y $z\n"
	}
	append out "POINT_DATA $_numPoints\n"
	set _vtkdata $out
	foreach axis {x y z} {
	set vector [set ${axis}v]
	    set _limits(${axis}min) [${vector} min]
	    set _limits(${axis}max) [${vector} max]
	}
    }
    blt::vector destroy $xv $yv $zv $all
}

itcl::body Rappture::Mesh::ReadVtk { xmlobj path } {
    set _type "vtk"
    if { $_dim < 1 } {
	error "can't read vtk mesh: no dimension specified"
    }

    # Write the contents to a file just in case it's binary.
    set tmpfile file[pid].vtk
    set f [open "$tmpfile" "w"]
    fconfigure $f -translation binary -encoding binary
    puts $f [$xmlobj get vtk]
    close $f

    # Read the data into a polydata output
    set _vtkreader $this-datasetreader
    vtkDataSetReader $_vtkreader
    $_vtkreader SetFileName $tmpfile
    $_vtkreader Update
    set _vtkoutput [$_vtkreader GetOutput]
    set _vtkpoints [$_vtkoutput GetPointData]
    file delete $tmpfile
}

itcl::body Rappture::Mesh::ReadGrid { xmlobj path } {
    set _type "grid"
    set numUniform 0
    set numNonUniform 0
    foreach axis { x y z } {
	set min    [$xmlobj get "$path.grid.${axis}axis.min"]
	set max    [$xmlobj get "$path.grid.${axis}axis.max"]
	set num    [$xmlobj get "$path.grid.${axis}axis.numpoints"]
	set values [$xmlobj get "$path.grid.${axis}values"]
	if { $min != "" && $max != "" && $num != "" && $num > 0 } {
	    set ${axis}Min $min
	    set ${axis}Max $max
	    set ${axis}Num $num
	    incr numUniform
	} elseif { $values != "" } {
	    incr numNonUniform
	    set ${axis}Values $values
	}
    }
    set _dim [expr $numNonUniform + $numUniform]
    if { $_dim == 0 } {
	# No data found.
	return 0
    }
    if { $numNonUniform == 0 && $numUniform > 0 } {
	# Specially handle the case where all axes are uniform.
	# This means we can use a vtkImageData to represent the data
	# instead of a vtkRectlinearGrid.
	# Vtk points are always 3D	
	if { $_dim == 2 } {
	    append out "DATASET STRUCTURED_POINTS\n"
	    append out "DIMENSIONS $xNum $yNum 1\n"
	    set xSpace [expr ($xMax - $xMin) / double($xNum - 1)]
	    set ySpace [expr ($yMax - $yMin) / double($yNum - 1)]
	    append out "SPACING $xSpace $ySpace 0\n"
	    append out "ORIGIN 0 0 0\n"
	    set _numPoints [expr $xNum * $yNum]
	    append out "POINT_DATA $_numPoints\n"
	    set _vtkdata $out
	    foreach axis {x y} {
		set _limits(${axis}min) [set ${axis}Min]
		set _limits(${axis}max) [set ${axis}Max]
	    }
	} elseif { $_dim == 3 } {
	    append out "DATASET STRUCTURED_POINTS\n"
	    append out "DIMENSIONS $xNum $yNum $zNum\n"
	    set xSpace [expr ($xMax - $xMin) / double($xNum - 1)]
	    set ySpace [expr ($yMax - $yMin) / double($yNum - 1)]
	    set zSpace [expr ($zMax - $zMin) / double($zNum - 1)]
	    append out "SPACING $xSpace $ySpace $zSpace\n"
	    append out "ORIGIN 0 0 0\n"
	    set _numPoints [expr $xNum * $yNum * zNum]
	    append out "POINT_DATA $_numPoints\n"
	    set _vtkdata $out
	    foreach axis {x y z} {
		set _limits(${axis}min) [set ${axis}Min]
		set _limits(${axis}max) [set ${axis}Max]
	    }
	} else { 
	    error "bad dimension of mesh \"$_dim\""
	}
	return 1
    }

    set xv [blt::vector create \#auto]
    if { [info exists xMin] } {
	$xv seq $xMin $xMax $xNum
    } else {
	$xv set [$xmlobj get $path.grid.xvalues]
	set xMin [$xv min]
	set xMax [$xv max]
	set xNum [$xv length]
    }
    set yv [blt::vector create \#auto]
    if { [info exists yMin] } {
	$yv seq $yMin $yMax $yNum
    } else {
	$yv set [$xmlobj get $path.grid.yvalues]
	set yMin [$yv min]
	set yMax [$yv max]
	set yNum [$yv length]
    }
    set zv [blt::vector create \#auto]
    if { $_dim == 3 } {
	if { [info exists zMin] } {
	    $zv seq $zMin $zMax $zNum
	}  else {
	    $zv set [$xmlobj get $path.grid.zvalues]
	    set zMin [$zv min]
	    set zMax [$zv max]
	    set zNum [$zv length]
	}
    } else {
	set zNum 1
    }
    if { $_dim == 3 } {
	set _numPoints [expr $xNum * $yNum * $zNum]
	append out "DATASET RECTILINEAR_GRID\n"
	append out "DIMENSIONS $xNum $yNum $zNum\n"
	append out "X_COORDINATES $xNum double\n"
	append out [$xv range 0 end]
	append out "\n"
	append out "Y_COORDINATES $yNum double\n"
	append out [$yv range 0 end]
	append out "\n"
	append out "Z_COORDINATES $zNum double\n"
	append out [$zv range 0 end]
	append out "\n"
	append out "POINT_DATA $_numPoints\n"
	set _vtkdata $out
	foreach axis {x y z} {
	    if { [info exists ${axis}Min] } {
		set _limits(${axis}min) [set ${axis}Min]
		set _limits(${axis}max) [set ${axis}Max]
	    }
	}
    } elseif { $_dim == 2 } {
	set _numPoints [expr $xNum * $yNum]
	append out "DATASET RECTILINEAR_GRID\n"
	append out "DIMENSIONS $xNum $yNum 1\n"
	append out "X_COORDINATES $xNum double\n"
	append out [$xv range 0 end]
	append out "\n"
	append out "Y_COORDINATES $yNum double\n"
	append out [$yv range 0 end]
	append out "\n"
	append out "Z_COORDINATES 1 double\n"
	append out "0\n"
	append out "POINT_DATA $_numPoints\n"
	foreach axis {x y} {
	    if { [info exists ${axis}Min] } {
		set _limits(${axis}min) [set ${axis}Min]
		set _limits(${axis}max) [set ${axis}Max]
	    }
	}
	set _vtkdata $out
    } else {
	error "invalid dimension of grid \"$_dim\""
    }
    blt::vector destroy $xv $yv $zv 
    return 1
}

itcl::body Rappture::Mesh::ReadTriangles { xmlobj path } {
    set _type "triangles"
    set _dim 2
    set triangles [$xmlobj get $path.triangles.indices]
    if { $triangles == "" } { 
	puts stderr "no triangle indices specified in mesh"
	return 0
    }
    set xvalues [$xmlobj get $path.triangles.xvalues]
    set yvalues [$xmlobj get $path.triangles.yvalues]
    set points [$xmlobj get $path.triangles.points]
    if { $xvalues != "" && $yvalues != "" } {
	set xv [blt::vector create \#auto]
	set yv [blt::vector create \#auto]
	$xv set $xvalues
	$yx set $yvalues
    } elseif { $points != "" } { 
	set all [blt::vector create \#auto]
	$all set $points
	set xv [blt::vector create \#auto]
	set yv [blt::vector create \#auto]
	$all split $xv $yv
    } else {
	puts stderr "missing either xvalues, yvalues, or points for triangular mesh"
	return 0
    }
    set _numPoints [$xv length]
    set count 0
    set data {}
    foreach { a b c } $triangles {
	append data "    3 $a $b $c\n"
	incr count
    }
    set celltypes {}
    for { set i 0 } { $i < $count } { incr i } {
	append celltypes "5\n"
    }
    append out "DATASET UNSTRUCTURED_GRID\n"
    append out "POINTS $_numPoints float\n"
    foreach x [$xv range 0 end] y [$yv range 0 end] {
	append out "    $x $y 1\n"
    }
    append out "CELLS $count [expr $count * 4]\n"
    append out $data
    append out "CELL_TYPES $count\n"
    append out $celltypes
    append out "POINT_DATA $_numPoints\n"
    foreach axis {x y} {
	if { [info exists ${axis}Min] } {
	    set _limits(${axis}min) [set ${axis}Min]
	    set _limits(${axis}max) [set ${axis}Max]
	}
    }
    set _vtkdata $out
}

itcl::body Rappture::Mesh::GetBltGridMesh { path } {
    if { $_dim != 2 } {
	error "must be 2D mesh for BLT"
    }
    # Vtk points are always 3D
    set xNum 0 
    if { $_xValues != "" } {
	blt::vector destroy $_xValues
    }
    set yNum 0
    if { $_yValues != "" } {
	blt::vector destroy $_yValues
    }
    set string [$xmlobj get grid.xvalues]
    set _xValues [blt::vector create \#auto]
    if { [llength $string] == 0 } {
	set xMin [GetDouble grid.xaxis.min]
	set xMax [GetDouble grid.xaxis.max]
	set xNum [GetInt grid.xaxis.numpoints]
	$xValues seq $xMin $xMax $xNum
    } else {
	$xValues set $string
	set xNum [$xValues length]
	set xMax [$xValues max]
	set xMin [$xValues min]
    }
    set _yValues [blt::vector create \#auto]
    set string [$xmlobj get grid.yvalues]
    if { [llength $string] == 0 } {
	set string [$xmlobj get grid.yaxis.min]
	if { $string != "" } {
	    set yMin [GetDouble grid.yaxis.min]
	    set yMax [GetDouble grid.yaxis.max]
	    set yNum [GetInt grid.yaxis.numpoints]
	    $yValues seq $yMin $yMax $yNum
	}
    } else {
	set yValues [blt::vector create \#auto]
	$yValues set $string 
	set yNum [$yValues length]
	set yMax [$yValues max]
	set yMin [$yValues min]
    }
    if { [$yValues length] == 0 } {
	# Not 2D
    }
    foreach axis {x y} {
	set _limits(${axis}min) [set ${axis}Min]
	set _limits(${axis}max) [set ${axis}Max]
    }
    set mesh [blt::mesh create irregular]
    blt::mesh configure $mesh -x $_xValues -y $_yValues
    return $mesh
}

itcl::body Rappture::Mesh::GetBltTriangleMesh { path } {
    set _xv [blt::vector create \#auto]
    set _yv [blt::vector create \#auto]
    $xv set [$_xmlobj get $path.triangle.xaxis.values]
    $yv set [$_xmlobj get $path.triangle.yaxis.values]
    set triangles [$_xmlobj get $path.triangle.triangles]
    set _limits(xmin) [$_xv min]
    set _limits(xmax) [$_xv max]
    set _limits(ymin) [$_yv min]
    set _limits(ymax) [$_yv max]
    set mesh [blt::mesh create triangle]
    blt::mesh configure $mesh -x $xv -y $yv -triangles $triangles
    return $mesh
}

itcl::body Rappture::Mesh::GetBltCloudMesh { path } {
    # Read the points into a vector and split them into individual coordinate
    # arrays.
    set all [blt::vector create \#auto]
    $all set [$_xmlobj get $path.cloud.points] 
    if { [$all length] & 0x1 } {
	error "odd number of points in 2D cloud mesh"
    }
    if { $_xv != "" } {
	blt::vector destroy $_xv
    }
    if { $_yv != "" } {
	blt::vector destroy $_yv
    }
    set _xv [blt::vector create \#auto]
    set _yv [blt::vector create \#auto]
    $all split $_xv $_yv
    set _limits(xmin) [$_xv min]
    set _limits(xmax) [$_xv max]
    set _limits(ymin) [$_yv min]
    set _limits(ymax) [$_yv max]
    set mesh [blt::mesh create cloud]
    blt::mesh configure $mesh -x $_xv -y $_yv 
    blt::vector destroy $all
    return $mesh
}

itcl::body Rappture::Mesh::GetBltMesh { path } {
    if { $_dim != 2 } {
	error "dimension of mesh $_dim is not 2"
    }
    switch -- $_type {
	"cloud" {
	    set _bltmesh [GetBltCloudMesh $path ]
	}
	"unirect" {
	    set _bltmesh [GetBltRegularMesh $path]
	}
	"rectangle" {
	    set _bltmesh [GetBltIrregularMesh $path]
	}
	"triangle" {
	    set _bltmesh [GetBltTriangleMesh $path]
	}
	default {
	    error "bad mesh type \"$_type\": should be cloud, unirect, rectangle, or triangle"
	}
    }
}

# ----------------------------------------------------------------------
# USAGE: ReadNodesElements <xmlobj> <path>
#
# Used internally to build a mesh representation based on nodes and
# elements stored in the XML.
# ----------------------------------------------------------------------
itcl::body Rappture::Mesh::ReadNodesElements {xmlobj path} {
    set type "nodeselements"
    # create the vtk objects containing points and connectivity
    set _dim 2 
    #
    # Extract each node and add it to the points list.
    #
    set points {}
    set count 0
    foreach comp [$xmlobj children -type node $path] {
        set xyz [$xmlobj get $path.$comp]

        # make sure we have x,y,z
        while {[llength $xyz] < 3} {
            lappend xyz "0"
        }

        # extract each point and add it to the points list
        foreach {x y z} $xyz break
        foreach dim {x y z} units $_units {
            set v [Rappture::Units::convert [set $dim] \
                -context $units -to $units -units off]

            set $dim $v  ;# save back to real x/y/z variable

            if {"" == $_limits(${dim}min)} {
                set _limits(${dim}min) $v
                set _limits(${dim}max) $v
            } else {
                if {$v < $_limits(${dim}min)} { set _limits(${dim}min) $v }
                if {$v > $_limits(${dim}max)} { set _limits(${dim}max) $v }
            }
        }
	append points "   $x $y $z\n"
	incr count
    }
    set _numPoints $count
    #
    # Extract each element and add it to the mesh.
    #
    array set node2celltype {
	3 5
	4 10
	8 11
	6 13
	5 14
    }
    set count 0
    set celltypes {}
    set data {}
    foreach comp [$xmlobj children -type element $path] {
        set nlist [$_mesh get $comp.nodes]
	set numNodes [llength $nlist]
	if { ![info exists node2celltype($numNodes)] } {
	    return 0
	}
	set celltype $node2celltype($numNodes)
	append celltypes "  $celltype\n"
	append data "  $numNodes $nlist\n"
	incr count
    }
    append out "DATASET UNSTRUCTURED_GRID\n"
    append out "POINTS $_numPoints float\n"
    append out $points
    append out "CELLS $count [expr $count * 4]\n"
    append out $data
    append out "CELL_TYPES $count\n"
    append out $celltypes
    append out "\n"
    append out "POINT_DATA $_numPoints\n"
    foreach axis {x y} {
	if { [info exists ${axis}Min] } {
	    set _limits(${axis}min) [set ${axis}Min]
	    set _limits(${axis}max) [set ${axis}Max]
	}
    }
    set _vtkdata $out
}
