# ----------------------------------------------------------------------
#  COMPONENT: vtkheightmapviewer - Vtk heightmap viewer
#
#  It connects to the Vtk server running on a rendering farm,
#  transmits data, and displays the results.
# ======================================================================
#  AUTHOR:  Michael McLennan, Purdue University
#  Copyright (c) 2004-2012  HUBzero Foundation, LLC
#
#  See the file "license.terms" for information on usage and
#  redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# ======================================================================
package require Itk
package require BLT
#package require Img

option add *VtkHeightmapViewer.width 4i widgetDefault
option add *VtkHeightmapViewer*cursor crosshair widgetDefault
option add *VtkHeightmapViewer.height 4i widgetDefault
option add *VtkHeightmapViewer.foreground black widgetDefault
option add *VtkHeightmapViewer.controlBackground gray widgetDefault
option add *VtkHeightmapViewer.controlDarkBackground #999999 widgetDefault
option add *VtkHeightmapViewer.plotBackground black widgetDefault
option add *VtkHeightmapViewer.plotForeground white widgetDefault
option add *VtkHeightmapViewer.font \
    -*-helvetica-medium-r-normal-*-12-* widgetDefault

# must use this name -- plugs into Rappture::resources::load
proc VtkHeightmapViewer_init_resources {} {
    Rappture::resources::register \
        vtkvis_server Rappture::VtkHeightmapViewer::SetServerList
}

itcl::class Rappture::VtkHeightmapViewer {
    inherit Rappture::VisViewer

    itk_option define -plotforeground plotForeground Foreground ""
    itk_option define -plotbackground plotBackground Background ""
    itk_option define -mode mode Mode "vtkcontour"

    constructor { hostlist args } {
        Rappture::VisViewer::constructor $hostlist
    } {
        # defined below
    }
    destructor {
        # defined below
    }
    public proc SetServerList { namelist } {
        Rappture::VisViewer::SetServerList "vtkvis" $namelist
    }
    public method add {dataobj {settings ""}}
    public method camera {option args}
    public method delete {args}
    public method disconnect {}
    public method download {option args}
    public method get {args}
    public method isconnected {}
    public method limits { colormap }
    public method sendto { string }
    public method parameters {title args} { 
        # do nothing 
    }
    public method scale {args}

    protected method Connect {}
    protected method CurrentDatasets {args}
    protected method Disconnect {}
    protected method DoResize {}
    protected method DoContour {}
    protected method DoRotate {}
    protected method AdjustSetting {what {value ""}}
    protected method InitSettings { args  }
    protected method Pan {option x y}
    protected method Pick {x y}
    protected method Rebuild {}
    protected method ReceiveDataset { args }
    protected method ReceiveImage { args }
    protected method ReceiveLegend { colormap title min max size }
    protected method Rotate {option x y}
    protected method SendCmd {string}
    protected method SendCmdNoWait {string}
    protected method Zoom {option}

    # The following methods are only used by this class.
    private method BuildAxisTab {}
    private method BuildCameraTab {}
    private method BuildColormap { name colors }
    private method GetColormap { color tag }
    private method BuildContourTab {}
    private method BuildDownloadPopup { widget command } 
    private method ChangeColormap { dataobj comp color }
    private method ColorsToColormap { color }
    private method Combo { option }
    private method ConvertToVtkData { dataobj comp } 
    private method DrawLegend { title }
    private method EnterLegend { x y } 
    private method EventuallyContour { numContours } 
    private method EventuallyResize { w h } 
    private method EventuallyRotate { q } 
    private method GetImage { args } 
    private method GetVtkData { args } 
    private method IsValidObject { dataobj } 
    private method LeaveLegend {}
    private method MotionLegend { x y } 
    private method PanCamera {}
    private method RequestLegend {}
    private method SetColormap { dataobj comp }
    private method SetLegendTip { x y }
    private method SetObjectStyle { dataobj comp } 
    private method GetHeightmapScale {} 
    private method EnterIsoline { x y value } 
    private method LeaveIsoline {}
    private method SetIsolineTip { x y value }

    private variable _arcball ""
    private variable _outbuf       ;    # buffer for outgoing commands

    private variable _dlist ""     ;    # list of data objects
    private variable _obj2datasets
    private variable _obj2ovride   ;    # maps dataobj => style override
    private variable _comp2scale;	# maps dataset to the heightmap scale.
    private variable _datasets     ;    # contains all the dataobj-component 
                                   ;    # datasets in the server
    private variable _colormaps    ;    # contains all the colormaps
                                   ;    # in the server.
    private variable _dataset2style    ;# maps dataobj-component to transfunc

    private variable _click        ;    # info used for rotate operations
    private variable _limits       ;    # autoscale min/max for all axes
    private variable _view         ;    # view params for 3D view
    private variable _settings
    private variable _style;            # Array of current component styles.
    private variable _initialStyle;     # Array of initial component styles.
    private variable _reset 1;          # indicates if camera needs to be reset
                                        # to starting position.

    private variable _first ""     ;    # This is the topmost dataset.
    private variable _start 0
    private variable _buffering 0
    private variable _title ""

    common _downloadPopup;              # download options from popup
    private common _hardcopy
    private variable _width 0
    private variable _height 0
    private variable _resizePending 0
    private variable _contourPending 0
    private variable _rotatePending 0
    private variable _legendPending 0
    private variable _outline
    private variable _fieldNames {} 
    private variable _fields 
    private variable _currentField "Default"
    private variable _numContours 0
    private variable _colorMode "vmag";#  Mode of colormap (vmag or scalar)
}

itk::usual VtkHeightmapViewer {
    keep -background -foreground -cursor -font
    keep -plotbackground -plotforeground -mode
}

# ----------------------------------------------------------------------
# CONSTRUCTOR
# ----------------------------------------------------------------------
itcl::body Rappture::VtkHeightmapViewer::constructor {hostlist args} {
    package require vtk
    set _serverType "vtkvis"

    # Rebuild event
    $_dispatcher register !rebuild
    $_dispatcher dispatch $this !rebuild "[itcl::code $this Rebuild]; list"

    # Resize event
    $_dispatcher register !resize
    $_dispatcher dispatch $this !resize "[itcl::code $this DoResize]; list"

    # Recontour event
    $_dispatcher register !contour
    $_dispatcher dispatch $this !contour "[itcl::code $this DoContour]; list"

    # Rotate event
    $_dispatcher register !rotate
    $_dispatcher dispatch $this !rotate "[itcl::code $this DoRotate]; list"

    # Legend event
    $_dispatcher register !legend
    $_dispatcher dispatch $this !legend "[itcl::code $this RequestLegend]; list"

    #
    # Populate parser with commands handle incoming requests
    #
    $_parser alias image [itcl::code $this ReceiveImage]
    $_parser alias dataset [itcl::code $this ReceiveDataset]
    $_parser alias legend [itcl::code $this ReceiveLegend]

    array set _outline {
        id -1
        afterId -1
        x1 -1
        y1 -1
        x2 -1
        y2 -1
    }
    # Initialize the view to some default parameters.
    array set _view {
	qw      0.36
	qx      0.25
	qy      0.50
	qz      0.70
        zoom	1.0 
        xpan	0
        ypan	0
        ortho	1
    }
    set _arcball [blt::arcball create 100 100]
    set q [list $_view(qw) $_view(qx) $_view(qy) $_view(qz)]
    $_arcball quaternion $q

    array set _settings [subst {
	debug-console		0
        axis-labels             1
        axis-visible            1
        axis-xgrid              0
        axis-xposition          0
        axis-ygrid              0
        axis-yposition          0
        axis-zgrid              0
        axis-zposition          0
	axis-minticks		1
	axis-flymode		"static"
        contour-edges           0
        contour-field           "Default"
        contour-isolines	10
        heightmap-lighting      1
        contour-opacity         100
        colormap-visible         1
        isolines-visible         1
        contour-wireframe       0
        heightmap-mode          0
        heightmap-scale         50
        legend-visible          1
	stretch-to-fit		0
	x-scale			1.0
	y-scale			1.0
	z-scale			1.0
    }]

    itk_component add view {
        canvas $itk_component(plotarea).view \
            -highlightthickness 0 -borderwidth 0
    } {
        usual
        ignore -highlightthickness -borderwidth -background 
    }

    itk_component add fieldmenu {
        menu $itk_component(plotarea).menu \
	    -relief flat \
            -tearoff no 
    } {
        usual
        ignore -background -foreground -relief -tearoff
    }
    set c $itk_component(view)
    bind $c <Configure> [itcl::code $this EventuallyResize %w %h]
    bind $c <4> [itcl::code $this Zoom in 0.25]
    bind $c <5> [itcl::code $this Zoom out 0.25]
    bind $c <KeyPress-Left>  [list %W xview scroll 10 units]
    bind $c <KeyPress-Right> [list %W xview scroll -10 units]
    bind $c <KeyPress-Up>    [list %W yview scroll 10 units]
    bind $c <KeyPress-Down>  [list %W yview scroll -10 units]
    bind $c <Enter> "focus %W"
    bind $c <Control-F1> [itcl::code $this ToggleConsole]

    # Fix the scrollregion in case we go off screen
    $c configure -scrollregion [$c bbox all]

    set _map(id) [$c create image 0 0 -anchor nw -image $_image(plot)]
    set _map(cwidth) -1 
    set _map(cheight) -1 
    set _map(zoom) 1.0
    set _map(original) ""

    set f [$itk_component(main) component controls]
    itk_component add reset {
        button $f.reset -borderwidth 1 -padx 1 -pady 1 \
            -highlightthickness 0 \
            -image [Rappture::icon reset-view] \
            -command [itcl::code $this Zoom reset]
    } {
        usual
        ignore -highlightthickness
    }
    pack $itk_component(reset) -side top -padx 2 -pady 2
    Rappture::Tooltip::for $itk_component(reset) "Reset the view to the default zoom level"

    itk_component add zoomin {
        button $f.zin -borderwidth 1 -padx 1 -pady 1 \
            -highlightthickness 0 \
            -image [Rappture::icon zoom-in] \
            -command [itcl::code $this Zoom in]
    } {
        usual
        ignore -highlightthickness
    }
    pack $itk_component(zoomin) -side top -padx 2 -pady 2
    Rappture::Tooltip::for $itk_component(zoomin) "Zoom in"

    itk_component add zoomout {
        button $f.zout -borderwidth 1 -padx 1 -pady 1 \
            -highlightthickness 0 \
            -image [Rappture::icon zoom-out] \
            -command [itcl::code $this Zoom out]
    } {
        usual
        ignore -highlightthickness
    }
    pack $itk_component(zoomout) -side top -padx 2 -pady 2
    Rappture::Tooltip::for $itk_component(zoomout) "Zoom out"

    itk_component add mode {
        Rappture::PushButton $f.mode \
            -onimage [Rappture::icon surface] \
            -offimage [Rappture::icon surface] \
            -variable [itcl::scope _settings(heightmap-mode)] \
            -command [itcl::code $this AdjustSetting heightmap-mode] \
    }
    Rappture::Tooltip::for $itk_component(mode) \
        "Toggle the surface/contour on/off"
    pack $itk_component(mode) -padx 2 -pady 2

    if { [catch {
        BuildContourTab
        BuildAxisTab
        BuildCameraTab
    } errs] != 0 } {
	global errorInfo
        puts stderr "errs=$errs errorInfo=$errorInfo"
    }
    set _image(legend) [image create photo]

    # Hack around the Tk panewindow.  The problem is that the requested 
    # size of the 3d view isn't set until an image is retrieved from
    # the server.  So the panewindow uses the tiny size.
    set w 10000
    pack forget $itk_component(view)
    blt::table $itk_component(plotarea) \
        0,0 $itk_component(view) -fill both -reqwidth $w 
    blt::table configure $itk_component(plotarea) c1 -resize none

    # Bindings for panning via mouse
    bind $itk_component(view) <ButtonPress-2> \
        [itcl::code $this Pan click %x %y]
    bind $itk_component(view) <B2-Motion> \
        [itcl::code $this Pan drag %x %y]
    bind $itk_component(view) <ButtonRelease-2> \
        [itcl::code $this Pan release %x %y]

    bind $itk_component(view) <ButtonRelease-3> \
        [itcl::code $this Pick %x %y]

    # Bindings for panning via keyboard
    bind $itk_component(view) <KeyPress-Left> \
        [itcl::code $this Pan set -10 0]
    bind $itk_component(view) <KeyPress-Right> \
        [itcl::code $this Pan set 10 0]
    bind $itk_component(view) <KeyPress-Up> \
        [itcl::code $this Pan set 0 -10]
    bind $itk_component(view) <KeyPress-Down> \
        [itcl::code $this Pan set 0 10]
    bind $itk_component(view) <Shift-KeyPress-Left> \
        [itcl::code $this Pan set -2 0]
    bind $itk_component(view) <Shift-KeyPress-Right> \
        [itcl::code $this Pan set 2 0]
    bind $itk_component(view) <Shift-KeyPress-Up> \
        [itcl::code $this Pan set 0 -2]
    bind $itk_component(view) <Shift-KeyPress-Down> \
        [itcl::code $this Pan set 0 2]

    # Bindings for zoom via keyboard
    bind $itk_component(view) <KeyPress-Prior> \
        [itcl::code $this Zoom out]
    bind $itk_component(view) <KeyPress-Next> \
        [itcl::code $this Zoom in]

    bind $itk_component(view) <Enter> "focus $itk_component(view)"

    if {[string equal "x11" [tk windowingsystem]]} {
        # Bindings for zoom via mouse
        bind $itk_component(view) <4> [itcl::code $this Zoom out]
        bind $itk_component(view) <5> [itcl::code $this Zoom in]
    }

    set _image(download) [image create photo]

    eval itk_initialize $args
    Connect
}

# ----------------------------------------------------------------------
# DESTRUCTOR
# ----------------------------------------------------------------------
itcl::body Rappture::VtkHeightmapViewer::destructor {} {
    Disconnect
    image delete $_image(plot)
    image delete $_image(download)
    catch { blt::arcball destroy $_arcball }
}

itcl::body Rappture::VtkHeightmapViewer::DoResize {} {
    if { $_width < 2 } {
        set _width 500
    }
    if { $_height < 2 } {
        set _height 500
    }
    set _start [clock clicks -milliseconds]
    SendCmd "screen size [expr $_width - 20] $_height"
    set _legendPending 1
    set _resizePending 0
}

itcl::body Rappture::VtkHeightmapViewer::DoRotate {} {
    set q [list $_view(qw) $_view(qx) $_view(qy) $_view(qz)]
    SendCmd "camera orient $q" 
    set _rotatePending 0
}

itcl::body Rappture::VtkHeightmapViewer::DoContour {} {
    foreach dataset [CurrentDatasets -all] {
        foreach {dataobj comp} [split $dataset -] break
        #SendCmd "heightmap numcontours $_numContours $dataset"
    }
    set _contourPending 0
}

itcl::body Rappture::VtkHeightmapViewer::EventuallyResize { w h } {
    set _width $w
    set _height $h
    $_arcball resize $w $h
    if { !$_resizePending } {
        set _resizePending 1
        $_dispatcher event -after 400 !resize
    }
}

itcl::body Rappture::VtkHeightmapViewer::EventuallyContour { numContours } {
    set _numContours $numContours
    puts stderr "Eventually _numContours=$_numContours"
    if { !$_contourPending } {
        set _contourPending 1
        $_dispatcher event -after 600 !contour
    }
}

set rotate_delay 100

itcl::body Rappture::VtkHeightmapViewer::EventuallyRotate { q } {
    foreach { _view(qw) _view(qx) _view(qy) _view(qz) } $q break
    if { !$_rotatePending } {
        set _rotatePending 1
        global rotate_delay 
        $_dispatcher event -after $rotate_delay !rotate
    }
}

# ----------------------------------------------------------------------
# USAGE: add <dataobj> ?<settings>?
#
# Clients use this to add a data object to the plot.  The optional
# <settings> are used to configure the plot.  Allowed settings are
# -color, -brightness, -width, -linestyle, and -raise.
# ----------------------------------------------------------------------
itcl::body Rappture::VtkHeightmapViewer::add {dataobj {settings ""}} {
    array set params {
        -color auto
        -width 1
        -linestyle solid
        -brightness 0
        -raise 0
        -description ""
        -param ""
        -type ""
    }
    array set params $settings
    set params(-description) ""
    set params(-param) ""
    foreach {opt val} $settings {
        if {![info exists params($opt)]} {
            error "bad setting \"$opt\": should be [join [lsort [array names params]] {, }]"
        }
        set params($opt) $val
    }
    if {$params(-color) == "auto" || $params(-color) == "autoreset"} {
        # can't handle -autocolors yet
        set params(-color) white
    }
    set pos [lsearch -exact $dataobj $_dlist]
    if {$pos < 0} {
        lappend _dlist $dataobj
    }
    set _obj2ovride($dataobj-color) $params(-color)
    set _obj2ovride($dataobj-width) $params(-width)
    set _obj2ovride($dataobj-raise) $params(-raise)
    $_dispatcher event -idle !rebuild
}


# ----------------------------------------------------------------------
# USAGE: delete ?<dataobj1> <dataobj2> ...?
#
#       Clients use this to delete a dataobj from the plot.  If no dataobjs
#       are specified, then all dataobjs are deleted.  No data objects are
#       deleted.  They are only removed from the display list.
#
# ----------------------------------------------------------------------
itcl::body Rappture::VtkHeightmapViewer::delete {args} {
    if { [llength $args] == 0} {
        set args $_dlist
    }
    # Delete all specified dataobjs
    set changed 0
    foreach dataobj $args {
        set pos [lsearch -exact $_dlist $dataobj]
        if { $pos < 0 } {
            continue;                   # Don't know anything about it.
        }
        # Remove it from the dataobj list.
        set _dlist [lreplace $_dlist $pos $pos]
        array unset _obj2ovride $dataobj-*
        array unset _settings $dataobj-*
        # Append to the end of the dataobj list.
        #lappend _dlist $dataobj
        set changed 1
    }
    # If anything changed, then rebuild the plot
    if { $changed } {
        $_dispatcher event -idle !rebuild
    }
}

# ----------------------------------------------------------------------
# USAGE: get ?-objects?
# USAGE: get ?-visible?
# USAGE: get ?-image view?
#
# Clients use this to query the list of objects being plotted, in
# order from bottom to top of this result.  The optional "-image"
# flag can also request the internal images being shown.
# ----------------------------------------------------------------------
itcl::body Rappture::VtkHeightmapViewer::get {args} {
    if {[llength $args] == 0} {
        set args "-objects"
    }

    set op [lindex $args 0]
    switch -- $op {
        "-objects" {
            # put the dataobj list in order according to -raise options
            set dlist {}
            foreach dataobj $_dlist {
                if { ![IsValidObject $dataobj] } {
                    continue
                }
                if {[info exists _obj2ovride($dataobj-raise)] && 
                    $_obj2ovride($dataobj-raise)} {
                    set dlist [linsert $dlist 0 $dataobj]
                } else {
                    lappend dlist $dataobj
                }
            }
            return $dlist
        }
        "-visible" {
            set dlist {}
            foreach dataobj $_dlist {
                if { ![IsValidObject $dataobj] } {
                    continue
                }
                if { ![info exists _obj2ovride($dataobj-raise)] } {
                    # No setting indicates that the object isn't visible.
                    continue
                }
                # Otherwise use the -raise parameter to put the object to
                # the front of the list.
                if { $_obj2ovride($dataobj-raise) } {
                    set dlist [linsert $dlist 0 $dataobj]
                } else {
                    lappend dlist $dataobj
                }
            }
            return $dlist
        }           
        -image {
            if {[llength $args] != 2} {
                error "wrong # args: should be \"get -image view\""
            }
            switch -- [lindex $args end] {
                view {
                    return $_image(plot)
                }
                default {
                    error "bad image name \"[lindex $args end]\": should be view"
                }
            }
        }
        default {
            error "bad option \"$op\": should be -objects or -image"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: scale ?<data1> <data2> ...?
#
# Sets the default limits for the overall plot according to the
# limits of the data for all of the given <data> objects.  This
# accounts for all objects--even those not showing on the screen.
# Because of this, the limits are appropriate for all objects as
# the user scans through data in the ResultSet viewer.
# ----------------------------------------------------------------------
itcl::body Rappture::VtkHeightmapViewer::scale {args} {
    foreach dataobj $args {
        set string [limits $dataobj]
        if { $string == "" } {
            continue
        }
        array set bounds $string
        if {![info exists _limits(xmin)] || $_limits(xmin) > $bounds(xmin)} {
            set _limits(xmin) $bounds(xmin)
        }
        if {![info exists _limits(xmax)] || $_limits(xmax) < $bounds(xmax)} {
            set _limits(xmax) $bounds(xmax)
        }

        if {![info exists _limits(ymin)] || $_limits(ymin) > $bounds(ymin)} {
            set _limits(ymin) $bounds(ymin)
        }
        if {![info exists _limits(ymax)] || $_limits(ymax) < $bounds(ymax)} {
            set _limits(ymax) $bounds(ymax)
        }

        if {![info exists _limits(zmin)] || $_limits(zmin) > $bounds(zmin)} {
            set _limits(zmin) $bounds(zmin)
        }
        if {![info exists _limits(zmax)] || $_limits(zmax) < $bounds(zmax)} {
            set _limits(zmax) $bounds(zmax)
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: download coming
# USAGE: download controls <downloadCommand>
# USAGE: download now
#
# Clients use this method to create a downloadable representation
# of the plot.  Returns a list of the form {ext string}, where
# "ext" is the file extension (indicating the type of data) and
# "string" is the data itself.
# ----------------------------------------------------------------------
itcl::body Rappture::VtkHeightmapViewer::download {option args} {
    switch $option {
        coming {
            if {[catch {
                blt::winop snap $itk_component(plotarea) $_image(download)
            }]} {
                $_image(download) configure -width 1 -height 1
                $_image(download) put #000000
            }
        }
        controls {
            set popup .vtkviewerdownload
            if { ![winfo exists .vtkviewerdownload] } {
                set inner [BuildDownloadPopup $popup [lindex $args 0]]
            } else {
                set inner [$popup component inner]
            }
            set _downloadPopup(image_controls) $inner.image_frame
            set num [llength [get]]
            set num [expr {($num == 1) ? "1 result" : "$num results"}]
            set word [Rappture::filexfer::label downloadWord]
            $inner.summary configure -text "$word $num in the following format:"
            update idletasks            ;# Fix initial sizes
            return $popup
        }
        now {
            set popup .vtkviewerdownload
            if {[winfo exists .vtkviewerdownload]} {
                $popup deactivate
            }
            switch -- $_downloadPopup(format) {
                "image" {
                    return [$this GetImage [lindex $args 0]]
                }
                "vtk" {
                    return [$this GetVtkData [lindex $args 0]]
                }
            }
            return ""
        }
        default {
            error "bad option \"$option\": should be coming, controls, now"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: Connect ?<host:port>,<host:port>...?
#
# Clients use this method to establish a connection to a new
# server, or to reestablish a connection to the previous server.
# Any existing connection is automatically closed.
# ----------------------------------------------------------------------
itcl::body Rappture::VtkHeightmapViewer::Connect {} {
    global readyForNextFrame
    set readyForNextFrame 1
    set _reset 1
    set _hosts [GetServerList "vtkvis"]
    if { "" == $_hosts } {
        return 0
    }
    set result [VisViewer::Connect $_hosts]
    if { $result } {
        set w [winfo width $itk_component(view)]
        set h [winfo height $itk_component(view)]
        EventuallyResize $w $h
    }
    return $result
}

#
# isconnected --
#
#       Indicates if we are currently connected to the visualization server.
#
itcl::body Rappture::VtkHeightmapViewer::isconnected {} {
    return [VisViewer::IsConnected]
}

#
# disconnect --
#
itcl::body Rappture::VtkHeightmapViewer::disconnect {} {
    Disconnect
    set _reset 1
}

#
# Disconnect --
#
#       Clients use this method to disconnect from the current rendering
#       server.
#
itcl::body Rappture::VtkHeightmapViewer::Disconnect {} {
    VisViewer::Disconnect

    $_dispatcher cancel !rebuild
    $_dispatcher cancel !resize
    $_dispatcher cancel !contour
    $_dispatcher cancel !rotate
    $_dispatcher cancel !legend
    # disconnected -- no more data sitting on server
    set _outbuf ""
    array unset _datasets 
    array unset _data 
    array unset _colormaps 
    array unset _dataset2style 
    array unset _obj2datasets 
    global readyForNextFrame
    set readyForNextFrame 1
}

#
# sendto --
#
itcl::body Rappture::VtkHeightmapViewer::sendto { bytes } {
    SendBytes "$bytes\n"
}

#
# SendCmd
#
#       Send commands off to the rendering server.  If we're currently
#       sending data objects to the server, buffer the commands to be 
#       sent later.
#
itcl::body Rappture::VtkHeightmapViewer::SendCmd {string} {
    if { $_buffering } {
        append _outbuf $string "\n"
    } else {
        SendBytes "$string\n"
    }
}

#
# SendCmdNoWait
#
#       Send commands off to the rendering server.  If we're currently
#       sending data objects to the server, buffer the commands to be 
#       sent later.
#
itcl::body Rappture::VtkHeightmapViewer::SendCmdNoWait {string} {
    if { $_buffering } {
        append _outbuf $string "\n"
    } else {
        SendBytes "$string\n"
    }
}

# ----------------------------------------------------------------------
# USAGE: ReceiveImage -bytes <size> -type <type> -token <token>
#
# Invoked automatically whenever the "image" command comes in from
# the rendering server.  Indicates that binary image data with the
# specified <size> will follow.
# ----------------------------------------------------------------------
itcl::body Rappture::VtkHeightmapViewer::ReceiveImage { args } {
    global readyForNextFrame
    set readyForNextFrame 1
    array set info {
        -token "???"
        -bytes 0
        -type image
    }
    array set info $args
    set bytes [ReceiveBytes $info(-bytes)]
    if { $info(-type) == "image" } {
        if 0 {
            set f [open "last.ppm" "w"] 
            puts $f $bytes
            close $f
        }
        $_image(plot) configure -data $bytes
        set time [clock seconds]
        set date [clock format $time]
        #puts stderr "$date: received image [image width $_image(plot)]x[image height $_image(plot)] image>"        
        if { $_start > 0 } {
            set finish [clock clicks -milliseconds]
            #puts stderr "round trip time [expr $finish -$_start] milliseconds"
            set _start 0
        }
    } elseif { $info(type) == "print" } {
        set tag $this-print-$info(-token)
        set _hardcopy($tag) $bytes
    }
    if { $_legendPending } {
        RequestLegend
    }
}

#
# ReceiveDataset --
#
itcl::body Rappture::VtkHeightmapViewer::ReceiveDataset { args } {
    if { ![isconnected] } {
        return
    }
    set option [lindex $args 0]
    switch -- $option {
        "scalar" {
            set option [lindex $args 1]
            switch -- $option {
                "world" {
                    foreach { x y z value tag } [lrange $args 2 end] break
                }
                "pixel" {
                    foreach { x y value tag } [lrange $args 2 end] break
                }
            }
        }
        "vector" {
            set option [lindex $args 1]
            switch -- $option {
                "world" {
                    foreach { x y z vx vy vz tag } [lrange $args 2 end] break
                }
                "pixel" {
                    foreach { x y vx vy vz tag } [lrange $args 2 end] break
                }
            }
        }
        "names" {
            foreach { name } [lindex $args 1] {
                #puts stderr "Dataset: $name"
            }
        }
        default {
            error "unknown dataset option \"$option\" from server"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: Rebuild
#
# Called automatically whenever something changes that affects the
# data in the widget.  Clears any existing data and rebuilds the
# widget to display new data.
# ----------------------------------------------------------------------
itcl::body Rappture::VtkHeightmapViewer::Rebuild {} {
    set w [winfo width $itk_component(view)]
    set h [winfo height $itk_component(view)]
    if { $w < 2 || $h < 2 } {
        $_dispatcher event -idle !rebuild
        return
    }

    set _buffering 1
    set _legendPending 1
    # Turn on buffering of commands to the server.  We don't want to
    # be preempted by a server disconnect/reconnect (which automatically
    # generates a new call to Rebuild).   
    if { $_width != $w || $_height != $h || $_reset } {
	set _width $w
	set _height $h
	$_arcball resize $w $h
	DoResize
	if { $_settings(stretch-to-fit) } {
	    Adjustsetting stretch-to-fit
	}
    }
    if { $_reset } {
 puts stderr "initial settings"
	InitSettings heightmap-mode background
	#
	# Reset the camera and other view parameters
	#
	SendCmd "axis color all [Color2RGB $itk_option(-plotforeground)]"
	SendCmd "axis lrot z 90"
	SendCmd "dataset maprange explicit $_limits(zmin) $_limits(zmax)"
	set q [list $_view(qw) $_view(qx) $_view(qy) $_view(qz)]
	$_arcball quaternion $q 
	if {$_settings(heightmap-mode) } {
	    SendCmd "camera reset all"
	    if { $_view(ortho)} {
		SendCmd "camera mode ortho"
	    } else {
		SendCmd "camera mode persp"
	    }
	} 	    
	DoRotate
	PanCamera
    }
    set _first ""

    #SendCmd "imgflush"

    set _first ""
    # Start off with no datasets are visible.
    SendCmd "dataset visible 0"
    set scale [GetHeightmapScale]
    foreach dataobj [get -objects] {
        if { [info exists _obj2ovride($dataobj-raise)] &&  $_first == "" } {
            set _first $dataobj
        }
	scale $dataobj
        set _obj2datasets($dataobj) ""
        foreach comp [$dataobj components] {
            set tag $dataobj-$comp
            if { ![info exists _datasets($tag)] } {
                set bytes [$dataobj vtkdata $comp]
		set f [open /tmp/vtkheightmap.vtk "w"]
		puts $f $bytes
		close $f
                set length [string length $bytes]
                append _outbuf "dataset add $tag data follows $length\n"
                append _outbuf $bytes
                set _datasets($tag) 1
                SetObjectStyle $dataobj $comp
            }
            lappend _obj2datasets($dataobj) $tag
            if { [info exists _obj2ovride($dataobj-raise)] } {
                SendCmd "heightmap visible 1 $tag"
            }
	    if { ![info exists _comp2scale($tag)] || 
		 $_comp2scale($tag) != $scale } {
		SendCmd "heightmap heightscale $scale $tag"
		set _comp2scale($tag) $scale
	    }
        }
    }
    if {"" != $_first} {
        set location [$_first hints camera]
        if { $location != "" } {
            array set view $location
        }

        if 1 {
            # Tell the server the name of the tool, the version, and dataset
            # that we are rendering.  Have to do it here because we don't know
            # what data objects are using the renderer until be get here.
            set args ""
            lappend args tool [$_first hints toolId]
            lappend args version [$_first hints toolRevision]
            lappend args dataset [$_first hints label]
            SendCmd "clientinfo [list $args]"
        }

	set _fieldNames [$_first hints fieldnames]
	set _fieldUnits [$_first hints fieldunits]
	set _fieldLabels [$_first hints fieldlabels]
        $itk_component(field) choices delete 0 end
        $itk_component(fieldmenu) delete 0 end
        array unset _fields
        foreach name $_fieldNames title $_fieldLabels units $_fieldUnits {
	    if { $title == "" } {
		set title $name
	    }
            $itk_component(field) choices insert end "$name" "$title"
            $itk_component(fieldmenu) add radiobutton -label "$title" \
                -value $title -variable [itcl::scope _currentField] \
                -selectcolor red \
                -activebackground $itk_option(-plotbackground) \
                -activeforeground $itk_option(-plotforeground) \
                -font "Arial 8" \
                -command [itcl::code $this Combo invoke]
            set _fields($name) [list $title $units]
	    set _currentField $name
        }
	if { [array size _fields] == 1 } {
	    set _currentField $_fieldLabels
	    if { $_currentField == "" } { 
		puts stderr "no default name from field"
		set _currentField "Default"
	    }
	}
	$itk_component(field) value $_currentField
    }

    InitSettings contour-colormap stretch-to-fit

    if { $_reset } {
	SendCmd "axis tickpos outside"
	SendCmd "axis range z $_limits(zmin) $_limits(zmax)"
	foreach axis { x y z } {
	    #SendCmd "axis range $axis $_limits(${axis}min) $_limits(${axis}max)"
	    SendCmd "axis lformat $axis %g"
	}
	foreach axis { x y z } {
	    set label [$_first hints ${axis}label]
	    if { $label == "" } {
		set label [string toupper $axis]
	    }
	    # May be a space in the axis label.
	    SendCmd [list axis name $axis $label]

	    set units [$_first hints ${axis}units]
	    if { $units != "" } {
		# May be a space in the axis units.
		SendCmd [list axis units $axis $units]
	    }
	}
	#
	# Reset the camera and other view parameters
	#
	SendCmd "axis color all [Color2RGB $itk_option(-plotforeground)]"
	SendCmd "dataset maprange all"
	set q [list $_view(qw) $_view(qx) $_view(qy) $_view(qz)]
	$_arcball quaternion $q 
	if {$_settings(heightmap-mode) } {
	    SendCmd "camera reset all"
	    if { $_view(ortho)} {
		SendCmd "camera mode ortho"
	    } else {
		SendCmd "camera mode persp"
	    }
	} 	    
	DoRotate
	PanCamera
	InitSettings axis-xgrid axis-ygrid axis-zgrid \
	    axis-visible axis-labels 
        InitSettings contour-opacity \
            heightmap-scale heightmap-lighting \
            contour-colormap contour-field \
            contour-edges contour-opacity contour-wireframe 
        set _reset 0
    }
    global readyForNextFrame
    set readyForNextFrame 0;		# Don't advance to the next frame
    set _buffering 0;			# Turn off buffering.

    # Actually write the commands to the server socket.  If it fails, we don't
    # care.  We're finished here.
    blt::busy hold $itk_component(hull)
    sendto $_outbuf;                        
    blt::busy release $itk_component(hull)
    set _outbuf "";                        # Clear the buffer.                
}

# ----------------------------------------------------------------------
# USAGE: CurrentDatasets ?-all -visible? ?dataobjs?
#
# Returns a list of server IDs for the current datasets being displayed.  This
# is normally a single ID, but it might be a list of IDs if the current data
# object has multiple components.
# ----------------------------------------------------------------------
itcl::body Rappture::VtkHeightmapViewer::CurrentDatasets {args} {
    set flag [lindex $args 0]
    switch -- $flag { 
        "-all" {
            if { [llength $args] > 1 } {
                error "CurrentDatasets: can't specify dataobj after \"-all\""
            }
            set dlist [get -objects]
        }
        "-visible" {
            if { [llength $args] > 1 } {
                set dlist {}
                set args [lrange $args 1 end]
                foreach dataobj $args {
                    if { [info exists _obj2ovride($dataobj-raise)] } {
                        lappend dlist $dataobj
                    }
                }
            } else {
                set dlist [get -visible]
            }
        }           
        default {
            set dlist $args
        }
    }
    set rlist ""
    foreach dataobj $dlist {
        foreach comp [$dataobj components] {
            set tag $dataobj-$comp
            if { [info exists _datasets($tag)] && $_datasets($tag) } {
                lappend rlist $tag
            }
        }
    }
    return $rlist
}

# ----------------------------------------------------------------------
# USAGE: Zoom in
# USAGE: Zoom out
# USAGE: Zoom reset
#
# Called automatically when the user clicks on one of the zoom
# controls for this widget.  Changes the zoom for the current view.
# ----------------------------------------------------------------------
itcl::body Rappture::VtkHeightmapViewer::Zoom {option} {
    switch -- $option {
        "in" {
            set _view(zoom) [expr {$_view(zoom)*1.25}]
            SendCmd "camera zoom $_view(zoom)"
        }
        "out" {
            set _view(zoom) [expr {$_view(zoom)*0.8}]
            SendCmd "camera zoom $_view(zoom)"
        }
        "reset" {
            array set _view {
		qw      0.36
		qx      0.25
		qy      0.50
		qz      0.70
                zoom    1.0
                xpan   0
                ypan   0
            }
            SendCmd "camera reset all"
            if { $_first != "" } {
                set location [$_first hints camera]
                if { $location != "" } {
                    array set _view $location
                }
            }
            set q [list $_view(qw) $_view(qx) $_view(qy) $_view(qz)]
            $_arcball quaternion $q
            DoRotate
        }
    }
}

itcl::body Rappture::VtkHeightmapViewer::PanCamera {} {
    set x $_view(xpan)
    set y $_view(ypan)
    SendCmd "camera pan $x $y"
}


# ----------------------------------------------------------------------
# USAGE: Rotate click <x> <y>
# USAGE: Rotate drag <x> <y>
# USAGE: Rotate release <x> <y>
#
# Called automatically when the user clicks/drags/releases in the
# plot area.  Moves the plot according to the user's actions.
# ----------------------------------------------------------------------
itcl::body Rappture::VtkHeightmapViewer::Rotate {option x y} {
    switch -- $option {
        "click" {
            $itk_component(view) configure -cursor fleur
            set _click(x) $x
            set _click(y) $y
        }
        "drag" {
            if {[array size _click] == 0} {
                Rotate click $x $y
            } else {
                set w [winfo width $itk_component(view)]
                set h [winfo height $itk_component(view)]
                if {$w <= 0 || $h <= 0} {
                    return
                }

                if {[catch {
                    # this fails sometimes for no apparent reason
                    set dx [expr {double($x-$_click(x))/$w}]
                    set dy [expr {double($y-$_click(y))/$h}]
                }]} {
                    return
                }
                if { $dx == 0 && $dy == 0 } {
                    return
                }
                set q [$_arcball rotate $x $y $_click(x) $_click(y)]
                EventuallyRotate $q
                set _click(x) $x
                set _click(y) $y
            }
        }
        "release" {
            Rotate drag $x $y
            $itk_component(view) configure -cursor ""
            catch {unset _click}
        }
        default {
            error "bad option \"$option\": should be click, drag, release"
        }
    }
}

itcl::body Rappture::VtkHeightmapViewer::Pick {x y} {
    foreach tag [CurrentDatasets -visible] {
        SendCmd "dataset getscalar pixel $x $y $tag"
    } 
}

# ----------------------------------------------------------------------
# USAGE: $this Pan click x y
#        $this Pan drag x y
#        $this Pan release x y
#
# Called automatically when the user clicks on one of the zoom
# controls for this widget.  Changes the zoom for the current view.
# ----------------------------------------------------------------------
itcl::body Rappture::VtkHeightmapViewer::Pan {option x y} {
    switch -- $option {
        "set" {
            set w [winfo width $itk_component(view)]
            set h [winfo height $itk_component(view)]
            set x [expr $x / double($w)]
            set y [expr $y / double($h)]
            set _view(xpan) [expr $_view(xpan) + $x]
            set _view(ypan) [expr $_view(ypan) + $y]
            PanCamera
            return
        }
        "click" {
            set _click(x) $x
            set _click(y) $y
            $itk_component(view) configure -cursor hand1
        }
        "drag" {
            if { ![info exists _click(x)] } {
                set _click(x) $x
            }
            if { ![info exists _click(y)] } {
                set _click(y) $y
            }
            set w [winfo width $itk_component(view)]
            set h [winfo height $itk_component(view)]
            set dx [expr ($_click(x) - $x)/double($w)]
            set dy [expr ($_click(y) - $y)/double($h)]
            set _click(x) $x
            set _click(y) $y
            set _view(xpan) [expr $_view(xpan) - $dx]
            set _view(ypan) [expr $_view(ypan) - $dy]
            PanCamera
        }
        "release" {
            Pan drag $x $y
            $itk_component(view) configure -cursor ""
        }
        default {
            error "unknown option \"$option\": should set, click, drag, or release"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: InitSettings <what> ?<value>?
#
# Used internally to update rendering settings whenever parameters
# change in the popup settings panel.  Sends the new settings off
# to the back end.
# ----------------------------------------------------------------------
itcl::body Rappture::VtkHeightmapViewer::InitSettings { args } {
    foreach spec $args {
        if { [info exists _settings($_first-$spec)] } {
            # Reset global setting with dataobj specific setting
            set _settings($spec) $_settings($_first-$spec)
        }
        AdjustSetting $spec
    }
}

#
# AdjustSetting --
#
#       Changes/updates a specific setting in the widget.  There are
#       usually user-setable option.  Commands are sent to the render
#       server.
#
itcl::body Rappture::VtkHeightmapViewer::AdjustSetting {what {value ""}} {
    if { ![isconnected] } {
        return
    }
    switch -- $what {
        "contour-opacity" {
            set val $_settings(contour-opacity)
            set sval [expr { 0.01 * double($val) }]
            foreach dataset [CurrentDatasets -all] {
                SendCmd "heightmap opacity $sval $dataset"
            }
        }
        "contour-wireframe" {
            set bool $_settings(contour-wireframe)
            foreach dataset [CurrentDatasets -all] {
                SendCmd "heightmap wireframe $bool $dataset"
            }
        }
        "colormap-visible" {
	    set bool $_settings(colormap-visible)
            foreach dataset [CurrentDatasets -all] {
                SendCmd "heightmap surface $bool $dataset"
            }
        }
        "isolines-visible" {
	    set bool $_settings(isolines-visible)
            foreach dataset [CurrentDatasets -all] {
                SendCmd "heightmap isolines $bool $dataset"
            }
        }
        "heightmap-lighting" {
	    if { $_settings(heightmap-mode) } {
		set bool $_settings(heightmap-lighting)
		SendCmd "heightmap lighting $bool"
	    } else {
		SendCmd "heightmap lighting 0"
	    }
        }
        "contour-edges" {
            set bool $_settings(contour-edges)
            foreach dataset [CurrentDatasets -all] {
                SendCmd "heightmap edges $bool $dataset"
            }
        }
        "axis-visible" {
            set bool $_settings(axis-visible)
            SendCmd "axis visible all $bool"
        }
        "axis-labels" {
            set bool $_settings(axis-labels)
            SendCmd "axis labels all $bool"
        }
        "axis-xgrid" - "axis-ygrid" - "axis-zgrid" {
            set axis [string range $what 5 5]
            set bool $_settings($what)
            SendCmd "axis grid $axis $bool"
        }
        "axis-flymode" {
            set mode [$itk_component(axis-flymode) value]
            set mode [$itk_component(axis-flymode) translate $mode]
            set _settings($what) $mode
            SendCmd "axis flymode $mode"
        }
        "axis-minticks" {
            set bool $_settings(axis-minticks)
	    foreach axis { x y z } { 
		SendCmd "axis minticks ${axis} $bool"
	    }
        }
        "contour-isolines" {
            set n $_settings($what)
            EventuallyContour $n
        }
        "isolines-visible" {
            set bool $_settings($what)
            foreach dataset [CurrentDatasets -all] {
                SendCmd "heightmap isolines $bool $dataset"
            }
        }
        "colormap-visible" {
            set bool $_settings($what)
            foreach dataset [CurrentDatasets -all] {
                SendCmd "heightmap surface $bool $dataset"
            }
        }
        "background" {
            set bgcolor [$itk_component(background) value]
 puts stderr "set background to $bgcolor"
	    array set fgcolors {
		"black" "white"
		"white" "black"
		"grey"	"black"
	    }
            configure -plotbackground $bgcolor \
		-plotforeground $fgcolors($bgcolor)
	    $itk_component(view) delete "legend"
	    RequestLegend
        }
        "isoline-color" {
            set color [$itk_component(isolinecolor) value]
	    if { $color == "none" } {
		if { $_settings(isolines-visible) } {
		    SendCmd "heightmap isolines 0"
		    set _settings(isolines-visible) 0
		}
	    } else {
		if { !$_settings(isolines-visible) } {
		    SendCmd "heightmap isolines 1"
		    set _settings(isolines-visible) 1
		}
		SendCmd "heightmap isolinecolor [Color2RGB $color]"
	    }
	    RequestLegend
        }
        "contour-colormap" {
            set color [$itk_component(colormap) value]
            set _settings(contour-colormap) $color
	    if { $color == "none" } {
		if { $_settings(colormap-visible) } {
		    SendCmd "heightmap surface 0"
		    set _settings(colormap-visible) 0
		}
	    } else {
		if { !$_settings(colormap-visible) } {
		    SendCmd "heightmap surface 1"
		    set _settings(colormap-visible) 1
		}
		set cmap [GetColormap $color "all"]
		if { ![info exists _dataset2style(all)] ||
		     $_dataset2style(all) != $cmap } {
		    SendCmd "heightmap colormap $cmap"
		    set _dataset2style(all) $cmap
		}
	    }
	    RequestLegend
        }
        "contour-opacity" {
            set val $_settings(contour-opacity)
            set sval [expr { 0.01 * double($val) }]
            foreach dataset [CurrentDatasets -all] {
                SendCmd "heightmap opacity $sval $dataset"
            }
        }
        "heightmap-scale" {
	    if { $_settings(heightmap-mode) } {
		set scale [GetHeightmapScale]
		foreach dataset [CurrentDatasets -all] {
		    SendCmd "heightmap heightscale $scale $dataset"
		    set _comp2scale($dataset) $scale
		}
	    }
        }
        "heightmap-lighting" {
	    if { $_settings(heightmap-mode) } {
		set bool $_settings(heightmap-lighting)
		SendCmd "heightmap lighting $bool"
	    } else {
		SendCmd "heightmap lighting 0"
	    }
        }
        "contour-field" {
            set new [$itk_component(field) value]
            set value [$itk_component(field) translate $new]
            set _settings(contour-field) $value
            if { [info exists _fields($new)] } {
                set _colorMode scalar
                set _currentField $new
            } else {
                puts stderr "unknown field \"$new\""
                return
            }
            set _legendPending 1
        }
        "stretch-to-fit" {
	    set bool $_settings(stretch-to-fit)
	    set _settings(x-scale) 1.0
	    set _settings(y-scale) 1.0
	    set _settings(z-scale) 1.0
	    if { $bool } {
		set w $_width
		set h $_height
		set xr [expr $_limits(xmax) - $_limits(xmin)]
		set yr [expr $_limits(ymax) - $_limits(ymin)]
		if { $h < $w } {
		    set xs [expr (double($w)*$yr) / (double($h)*$xr)]
		    set _settings(x-scale) $xs
		} else {
		    set ys [expr (double($h)*$xr) / (double($w)*$yr)]
		    set _settings(y-scale) $ys
		    set _settings(z-scale) $ys
		}
	    }
	    SendCmd "heightmap scale $_settings(x-scale) $_settings(y-scale) $_settings(z-scale)"
	    # It kind of makes sense that I need to reset the camera 
	    # whenever the scale changes.
	    Zoom reset
	}	    
        "heightmap-mode" {
	    set bool $_settings(heightmap-mode)
	    if { $bool } {
		$itk_component(lighting) configure -state normal
	    } else {
		$itk_component(lighting) configure -state disabled
	    }
            if {[isconnected]} {
		set _settings(heightmap-lighting) $bool
		# Fix heightmap scale: 0 for contours, 1 for heightmaps.
		if { $bool } {
		    set _settings(heightmap-scale) 50
		} else {
		    set _settings(heightmap-scale) 50
		}
                SendCmd "heightmap lighting $bool"
		set scale [GetHeightmapScale]
                foreach dataset [CurrentDatasets -all] {
		    SendCmd "heightmap heightscale $scale $dataset"
		    set _comp2scale($dataset) $scale
                }
		# Fix the mouse bindings for rotation/panning and the 
		# camera mode. Ideally we'd create a bindtag for these.
		set c $itk_component(view)
		if { $bool } {
		    # Bindings for rotation via mouse
		    bind $c <ButtonPress-1> \
			[itcl::code $this Rotate click %x %y]
		    bind $c <B1-Motion> \
			[itcl::code $this Rotate drag %x %y]
		    bind $c <ButtonRelease-1> \
			[itcl::code $this Rotate release %x %y]
		    set q [list $_view(qw) $_view(qx) $_view(qy) $_view(qz)]
		    $_arcball quaternion $q
                    if {$_view(ortho)} {
                        SendCmd "camera mode ortho"
                    } else {
                        SendCmd "camera mode persp"
                    }
		    SendCmd "camera orient $q" 
		} else {
		    set _settings(heightmap-scale) 0
		    bind $c <ButtonPress-1> {}
		    bind $c <B1-Motion> {}
		    bind $c <ButtonRelease-1> {}
		    SendCmd "camera mode image"
		}
            }
        }        
        "legend-visible" {
            if { !$_settings(legend-visible) } {
		$itk_component(view) delete legend
	    }
	    RequestLegend
        }
	default {
            error "don't know how to fix $what"
        }
    }
}

#
# RequestLegend --
#
#       Request a new legend from the server.  The size of the legend
#       is determined from the height of the canvas.  It will be rotated
#       to be vertical when drawn.
#
itcl::body Rappture::VtkHeightmapViewer::RequestLegend {} {
    set font "Arial 8"
    set lineht [font metrics $font -linespace]
    set c $itk_component(view)
    set w 12
    set h [expr {$_height - 2 * ($lineht + 2)}]
    if { $h < 1} {
        return
    }
    if { [info exists _fields($_currentField)] } {
        set title [lindex $_fields($_currentField) 0]
	if { $title != "component" } {
	    set h [expr $h - ($lineht + 2)]
	}
    } else {
        return
    }
    # Set the legend on the first heightmap dataset.
    if { [info exists _dataset2style(all)] } {
	set cmap $_dataset2style(all)
	SendCmdNoWait "legend $cmap scalar $title {} $w $h 0"
    }
}

#
# ChangeColormap --
#
itcl::body Rappture::VtkHeightmapViewer::ChangeColormap {dataobj comp color} {
    set tag $dataobj-$comp
    if { ![info exist _style($tag)] } {
        error "no initial colormap"
    }
    array set style $_style($tag)
    set style(-color) $color
    set _style($tag) [array get style]
    SetColormap $dataobj $comp
}

#
# SetColormap --
#
itcl::body Rappture::VtkHeightmapViewer::SetColormap { dataobj comp } {
    array set style {
        -color BCGYR
        -levels 10
        -opacity 1.0
    }
    set tag $dataobj-$comp
    array set style $_style($tag)

    set name "$style(-color):$style(-levels):$style(-opacity)"
    if { ![info exists _colormaps($name)] } {
        BuildColormap $name [array get style]
        set _colormaps($name) 1
    }
    if { ![info exists _dataset2style($tag)] ||
         $_dataset2style($tag) != $name } {
        SendCmd "heightmap colormap $name $tag"
        set _dataset2style($tag) $name
    }
}

itcl::body Rappture::VtkHeightmapViewer::ColorsToColormap { colors } {
    switch -- $colors {
        "grey-to-blue" {
            return {
                0.0                      0.200 0.200 0.200
                0.14285714285714285      0.400 0.400 0.400
                0.2857142857142857       0.600 0.600 0.600
                0.42857142857142855      0.900 0.900 0.900
                0.5714285714285714       0.800 1.000 1.000
                0.7142857142857143       0.600 1.000 1.000
                0.8571428571428571       0.400 0.900 1.000
                1.0                      0.000 0.600 0.800
            }
        }
        "blue-to-grey" {
            return {
                0.0                     0.000 0.600 0.800 
                0.14285714285714285     0.400 0.900 1.000 
                0.2857142857142857      0.600 1.000 1.000 
                0.42857142857142855     0.800 1.000 1.000 
                0.5714285714285714      0.900 0.900 0.900 
                0.7142857142857143      0.600 0.600 0.600 
                0.8571428571428571      0.400 0.400 0.400 
                1.0                     0.200 0.200 0.200
            }
        }
        "blue" {
            return { 
                0.0                     0.900 1.000 1.000 
                0.1111111111111111      0.800 0.983 1.000 
                0.2222222222222222      0.700 0.950 1.000 
                0.3333333333333333      0.600 0.900 1.000 
                0.4444444444444444      0.500 0.833 1.000 
                0.5555555555555556      0.400 0.750 1.000 
                0.6666666666666666      0.300 0.650 1.000 
                0.7777777777777778      0.200 0.533 1.000 
                0.8888888888888888      0.100 0.400 1.000 
                1.0                     0.000 0.250 1.000
            }
        }
        "brown-to-blue" {
            return {
                0.0                             0.200   0.100   0.000 
                0.09090909090909091             0.400   0.187   0.000 
                0.18181818181818182             0.600   0.379   0.210 
                0.2727272727272727              0.800   0.608   0.480 
                0.36363636363636365             0.850   0.688   0.595 
                0.45454545454545453             0.950   0.855   0.808 
                0.5454545454545454              0.800   0.993   1.000 
                0.6363636363636364              0.600   0.973   1.000 
                0.7272727272727273              0.400   0.940   1.000 
                0.8181818181818182              0.200   0.893   1.000 
                0.9090909090909091              0.000   0.667   0.800 
                1.0                             0.000   0.480   0.600 
            }
        }
        "blue-to-brown" {
            return {
                0.0                             0.000   0.480   0.600 
                0.09090909090909091             0.000   0.667   0.800 
                0.18181818181818182             0.200   0.893   1.000 
                0.2727272727272727              0.400   0.940   1.000 
                0.36363636363636365             0.600   0.973   1.000 
                0.45454545454545453             0.800   0.993   1.000 
                0.5454545454545454              0.950   0.855   0.808 
                0.6363636363636364              0.850   0.688   0.595 
                0.7272727272727273              0.800   0.608   0.480 
                0.8181818181818182              0.600   0.379   0.210 
                0.9090909090909091              0.400   0.187   0.000 
                1.0                             0.200   0.100   0.000 
            }
        }
        "blue-to-orange" {
            return {
                0.0                             0.000   0.167   1.000
                0.09090909090909091             0.100   0.400   1.000
                0.18181818181818182             0.200   0.600   1.000
                0.2727272727272727              0.400   0.800   1.000
                0.36363636363636365             0.600   0.933   1.000
                0.45454545454545453             0.800   1.000   1.000
                0.5454545454545454              1.000   1.000   0.800
                0.6363636363636364              1.000   0.933   0.600
                0.7272727272727273              1.000   0.800   0.400
                0.8181818181818182              1.000   0.600   0.200
                0.9090909090909091              1.000   0.400   0.100
                1.0                             1.000   0.167   0.000
            }
        }
        "orange-to-blue" {
            return {
                0.0                             1.000   0.167   0.000
                0.09090909090909091             1.000   0.400   0.100
                0.18181818181818182             1.000   0.600   0.200
                0.2727272727272727              1.000   0.800   0.400
                0.36363636363636365             1.000   0.933   0.600
                0.45454545454545453             1.000   1.000   0.800
                0.5454545454545454              0.800   1.000   1.000
                0.6363636363636364              0.600   0.933   1.000
                0.7272727272727273              0.400   0.800   1.000
                0.8181818181818182              0.200   0.600   1.000
                0.9090909090909091              0.100   0.400   1.000
                1.0                             0.000   0.167   1.000
            }
        }
        "rainbow" {
            set clist {
                "#EE82EE"
                "#4B0082" 
                "blue" 
                "#008000" 
                "yellow" 
                "#FFA500" 
                "red" 
            }
        }
        "BGYOR" {
            set clist {
                "blue" 
                "#008000" 
                "yellow" 
                "#FFA500" 
                "red" 
            }
        }
        "ROYGB" {
            set clist {
                "red" 
                "#FFA500" 
                "yellow" 
                "#008000" 
                "blue" 
            }
        }
        "RYGCB" {
            set clist {
                "red" 
                "yellow" 
                "green"
                "cyan"
                "blue"
            }
        }
        "BCGYR" {
            set clist {
                "blue" 
                "cyan"
                "green"
                "yellow" 
                "red" 
            }
        }
        "spectral" {
            return {
                0.0 0.150 0.300 1.000 
                0.1 0.250 0.630 1.000 
                0.2 0.450 0.850 1.000 
                0.3 0.670 0.970 1.000 
                0.4 0.880 1.000 1.000 
                0.5 1.000 1.000 0.750 
                0.6 1.000 0.880 0.600 
                0.7 1.000 0.680 0.450 
                0.8 0.970 0.430 0.370 
                0.9 0.850 0.150 0.196 
                1.0 0.650 0.000 0.130
            }
        }
        "green-to-magenta" {
            return {
                0.0 0.000 0.316 0.000 
                0.06666666666666667 0.000 0.526 0.000 
                0.13333333333333333 0.000 0.737 0.000 
                0.2 0.000 0.947 0.000 
                0.26666666666666666 0.316 1.000 0.316 
                0.3333333333333333 0.526 1.000 0.526 
                0.4 0.737 1.000 0.737 
                0.4666666666666667 1.000 1.000 1.000 
                0.5333333333333333 1.000 0.947 1.000 
                0.6 1.000 0.737 1.000 
                0.6666666666666666 1.000 0.526 1.000 
                0.7333333333333333 1.000 0.316 1.000 
                0.8 0.947 0.000 0.947 
                0.8666666666666667 0.737 0.000 0.737 
                0.9333333333333333 0.526 0.000 0.526 
                1.0 0.316 0.000 0.316
            }
        }
        "greyscale" {
            return { 
                0.0 0.0 0.0 0.0 1.0 1.0 1.0 1.0
            }
        }
        "nanohub" {
            set clist "white yellow green cyan blue magenta"
        }
        default {
            set clist $colors
        }
    }
    set cmap {}
    for {set i 0} {$i < [llength $clist]} {incr i} {
        set x [expr {double($i)/([llength $clist]-1)}]
        set color [lindex $clist $i]
        append cmap "$x [Color2RGB $color] "
    }
puts stderr "colormap $colors = ($cmap)"
    return $cmap
}

#
# BuildColormap --
#
itcl::body Rappture::VtkHeightmapViewer::BuildColormap { name styles } {
    array set style $styles
    set cmap [ColorsToColormap $style(-color)]
    if { [llength $cmap] == 0 } {
        set cmap "0.0 0.0 0.0 0.0 1.0 1.0 1.0 1.0"
    }
    if { ![info exists _settings(contour-opacity)] } {
        set _settings(contour-opacity) $style(-opacity)
    }
    set max $_settings(contour-opacity)

    set wmap "0.0 1.0 1.0 1.0"
    SendCmd "colormap add $name { $cmap } { $wmap }"
}

# ----------------------------------------------------------------------
# CONFIGURATION OPTION: -mode
# ----------------------------------------------------------------------
itcl::configbody Rappture::VtkHeightmapViewer::mode {
    switch -- $itk_option(-mode) {
	"vtkheightmap" {
	    set _settings(heightmap-mode) 1
	}
	"vtkcontour" {
	    set _settings(heightmap-mode) 0
	} 
	default {
	    error "unknown mode settings \"$itk_option(-mode)\""
	}
    }
    AdjustSetting heightmap-mode
}

# ----------------------------------------------------------------------
# CONFIGURATION OPTION: -plotbackground
# ----------------------------------------------------------------------
itcl::configbody Rappture::VtkHeightmapViewer::plotbackground {
    if { [isconnected] } {
        SendCmd "screen bgcolor [Color2RGB $itk_option(-plotbackground)]"
	$itk_component(view) configure -background $itk_option(-plotbackground)
    }
}

# ----------------------------------------------------------------------
# CONFIGURATION OPTION: -plotforeground
# ----------------------------------------------------------------------
itcl::configbody Rappture::VtkHeightmapViewer::plotforeground {
    if { [isconnected] } {
	SendCmd "axis color all [Color2RGB $itk_option(-plotforeground)]"
    }
}

itcl::body Rappture::VtkHeightmapViewer::limits { dataobj } {
    foreach { limits(xmin) limits(xmax) } [$dataobj limits x] break
    foreach { limits(ymin) limits(ymax) } [$dataobj limits y] break
    foreach { limits(zmin) limits(zmax) } [$dataobj limits v] break
    return [array get limits]
}

itcl::body Rappture::VtkHeightmapViewer::BuildContourTab {} {

    set fg [option get $itk_component(hull) font Font]
    #set bfg [option get $itk_component(hull) boldFont Font]

    set inner [$itk_component(main) insert end \
        -title "Contour Settings" \
        -icon [Rappture::icon volume-on]]
    $inner configure -borderwidth 4

    checkbutton $inner.legend \
        -text "Legend" \
        -variable [itcl::scope _settings(legend-visible)] \
        -command [itcl::code $this AdjustSetting legend-visible] \
        -font "Arial 9"

    checkbutton $inner.wireframe \
        -text "Wireframe" \
        -variable [itcl::scope _settings(contour-wireframe)] \
        -command [itcl::code $this AdjustSetting contour-wireframe] \
        -font "Arial 9"

    itk_component add lighting {
	checkbutton $inner.lighting \
	    -text "Enable Lighting" \
	    -variable [itcl::scope _settings(heightmap-lighting)] \
	    -command [itcl::code $this AdjustSetting heightmap-lighting] \
	    -font "Arial 9"
    }

    checkbutton $inner.edges \
        -text "Edges" \
        -variable [itcl::scope _settings(contour-edges)] \
        -command [itcl::code $this AdjustSetting contour-edges] \
        -font "Arial 9"

    checkbutton $inner.stretch \
        -text "Stretch to fit" \
        -variable [itcl::scope _settings(stretch-to-fit)] \
        -command [itcl::code $this AdjustSetting stretch-to-fit] \
        -font "Arial 9"

    label $inner.field_l -text "Field" -font "Arial 9" 
    itk_component add field {
        Rappture::Combobox $inner.field -width 10 -editable no
    }
    bind $inner.field <<Value>> \
        [itcl::code $this AdjustSetting contour-field]

    label $inner.colormap_l -text "Colormap" -font "Arial 9" 
    itk_component add colormap {
        Rappture::Combobox $inner.colormap -width 10 -editable no
    }
    $inner.colormap choices insert end \
        "BCGYR"              "BCGYR"            \
        "BGYOR"              "BGYOR"            \
        "blue"               "blue"             \
        "blue-to-brown"      "blue-to-brown"    \
        "blue-to-orange"     "blue-to-orange"   \
        "blue-to-grey"       "blue-to-grey"     \
        "green-to-magenta"   "green-to-magenta" \
        "greyscale"          "greyscale"        \
        "nanohub"            "nanohub"          \
        "rainbow"            "rainbow"          \
        "spectral"           "spectral"         \
        "ROYGB"              "ROYGB"            \
        "RYGCB"              "RYGCB"            \
        "brown-to-blue"      "brown-to-blue"    \
        "grey-to-blue"       "grey-to-blue"     \
        "orange-to-blue"     "orange-to-blue"   \
	"none"		     "none"

    $itk_component(colormap) value "BCGYR"
    bind $inner.colormap <<Value>> \
        [itcl::code $this AdjustSetting contour-colormap]

    label $inner.isolinecolor_l -text "Isolines" -font "Arial 9" 
    itk_component add isolinecolor {
        Rappture::Combobox $inner.isolinecolor -width 10 -editable no
    }
    $inner.isolinecolor choices insert end \
        "black"              "black"            \
        "blue"               "blue"             \
        "cyan"               "cyan"             \
        "green"              "green"            \
        "grey"               "grey"             \
        "magenta"            "magenta"          \
        "orange"             "orange"           \
        "red"                "red"              \
        "white"              "white"            \
	"none"		     "none"

    $itk_component(isolinecolor) value "red"
    bind $inner.isolinecolor <<Value>> \
	[itcl::code $this AdjustSetting isoline-color]

    label $inner.background_l -text "Background" -font "Arial 9" 
    itk_component add background {
        Rappture::Combobox $inner.background -width 10 -editable no
    }
    $inner.background choices insert end \
        "black"              "black"            \
        "white"              "white"            \
        "grey"               "grey"             

    $itk_component(background) value "white"
    bind $inner.background <<Value>> [itcl::code $this AdjustSetting background]

    label $inner.opacity_l -text "Opacity" -font "Arial 9"
    ::scale $inner.opacity -from 0 -to 100 -orient horizontal \
        -variable [itcl::scope _settings(contour-opacity)] \
        -width 10 \
        -showvalue off \
        -command [itcl::code $this AdjustSetting contour-opacity]

    label $inner.scale_l -text "Scale" -font "Arial 9"
    ::scale $inner.scale -from 0 -to 100 -orient horizontal \
        -variable [itcl::scope _settings(heightmap-scale)] \
        -width 10 \
        -showvalue off \
        -command [itcl::code $this AdjustSetting heightmap-scale]

    blt::table $inner \
        0,0 $inner.colormap_l -anchor w -pady 2  \
        0,1 $inner.colormap   -anchor w -pady 2 -fill x  \
        1,0 $inner.isolinecolor_l  -anchor w -pady 2  \
        1,1 $inner.isolinecolor    -anchor w -pady 2 -fill x  \
	2,0 $inner.background_l -anchor w -pady 2 \
	2,1 $inner.background -anchor w -pady 2  -fill x \
        4,0 $inner.legend    -anchor w -pady 2 -cspan 2 \
        5,0 $inner.wireframe -anchor w -pady 2 -cspan 2\
        6,0 $inner.lighting  -anchor w -pady 2 -cspan 2 \
        7,0 $inner.edges     -anchor w -pady 2 -cspan 2 \
        8,0 $inner.stretch     -anchor w -pady 2 -cspan 2 \
        9,0 $inner.opacity_l -anchor w -pady 2 \
        9,1 $inner.opacity   -fill x   -pady 2 \
        10,0 $inner.scale_l   -anchor w -pady 2 -cspan 2 \
        10,1 $inner.scale     -fill x   -pady 2 -cspan 2 \

    if { [array size _fields] > 1 } {
	blt::table $inner \
	    3,0 $inner.field_l   -anchor w -pady 2 \
	    3,1 $inner.field     -anchor w -pady 2 -fill x 
    }
    blt::table configure $inner r* c* -resize none
    blt::table configure $inner r11 c1 -resize expand
}

itcl::body Rappture::VtkHeightmapViewer::BuildAxisTab {} {

    set fg [option get $itk_component(hull) font Font]
    #set bfg [option get $itk_component(hull) boldFont Font]

    set inner [$itk_component(main) insert end \
        -title "Axis Settings" \
        -icon [Rappture::icon axis1]]
    $inner configure -borderwidth 4

    checkbutton $inner.visible \
        -text "Axes" \
        -variable [itcl::scope _settings(axis-visible)] \
        -command [itcl::code $this AdjustSetting axis-visible] \
        -font "Arial 9"
    checkbutton $inner.labels \
        -text "Axis Labels" \
        -variable [itcl::scope _settings(axis-labels)] \
        -command [itcl::code $this AdjustSetting axis-labels] \
        -font "Arial 9"
    label $inner.grid_l -text "Grid" -font "Arial 9" 
    checkbutton $inner.gridx \
        -text "X" \
        -variable [itcl::scope _settings(axis-xgrid)] \
        -command [itcl::code $this AdjustSetting axis-xgrid] \
        -font "Arial 9"
    checkbutton $inner.gridy \
        -text "Y" \
        -variable [itcl::scope _settings(axis-ygrid)] \
        -command [itcl::code $this AdjustSetting axis-ygrid] \
        -font "Arial 9"
    checkbutton $inner.gridz \
        -text "Z" \
        -variable [itcl::scope _settings(axis-zgrid)] \
        -command [itcl::code $this AdjustSetting axis-zgrid] \
        -font "Arial 9"
    checkbutton $inner.minorticks \
        -text "Minor Ticks" \
        -variable [itcl::scope _settings(axis-minticks)] \
        -command [itcl::code $this AdjustSetting axis-minticks] \
        -font "Arial 9"


    label $inner.mode_l -text "Mode" -font "Arial 9" 

    itk_component add axis-flymode {
        Rappture::Combobox $inner.mode -width 10 -editable no
    }
    $inner.mode choices insert end \
        "static_triad"    "static" \
        "closest_triad"   "closest" \
        "furthest_triad"  "furthest" \
        "outer_edges"     "outer"         
    $itk_component(axis-flymode) value "static"
    bind $inner.mode <<Value>> [itcl::code $this AdjustSetting axis-flymode]

    blt::table $inner \
        0,0 $inner.visible -anchor w -cspan 4 \
        1,0 $inner.labels  -anchor w -cspan 4 \
        2,0 $inner.minorticks  -anchor w -cspan 4 \
	4,0 $inner.grid_l  -anchor w \
        4,1 $inner.gridx   -anchor w \
        4,2 $inner.gridy   -anchor w \
        4,3 $inner.gridz   -anchor w \
        5,0 $inner.mode_l  -anchor w -padx { 2 0 } \
        5,1 $inner.mode    -fill x -cspan 3 

    blt::table configure $inner r* c* -resize none
    blt::table configure $inner r7 c6 -resize expand
    blt::table configure $inner r3 -height 0.125i
}


itcl::body Rappture::VtkHeightmapViewer::BuildCameraTab {} {
    set inner [$itk_component(main) insert end \
        -title "Camera Settings" \
        -icon [Rappture::icon camera]]
    $inner configure -borderwidth 4

    set labels { qx qy qz qw xpan ypan zoom }
    set row 0
    foreach tag $labels {
        label $inner.${tag}label -text $tag -font "Arial 9"
        entry $inner.${tag} -font "Arial 9"  -bg white \
            -textvariable [itcl::scope _view($tag)]
        bind $inner.${tag} <KeyPress-Return> \
            [itcl::code $this camera set ${tag}]
        blt::table $inner \
            $row,0 $inner.${tag}label -anchor e -pady 2 \
            $row,1 $inner.${tag} -anchor w -pady 2
        blt::table configure $inner r$row -resize none
        incr row
    }
    checkbutton $inner.ortho \
        -text "Orthographic Projection" \
        -variable [itcl::scope _view(ortho)] \
        -command [itcl::code $this camera set ortho] \
        -font "Arial 9"
    blt::table $inner \
            $row,0 $inner.ortho -columnspan 2 -anchor w -pady 2
    blt::table configure $inner r$row -resize none
    incr row

    blt::table configure $inner c0 c1 -resize none
    blt::table configure $inner c2 -resize expand
    blt::table configure $inner r$row -resize expand
}

#
#  camera -- 
#
itcl::body Rappture::VtkHeightmapViewer::camera {option args} {
    switch -- $option { 
        "show" {
            puts [array get _view]
        }
        "set" {
            set who [lindex $args 0]
            set x $_view($who)
            set code [catch { string is double $x } result]
            if { $code != 0 || !$result } {
                return
            }
            switch -- $who {
                "ortho" {
                    if {$_view(ortho)} {
                        SendCmd "camera mode ortho"
                    } else {
                        SendCmd "camera mode persp"
                    }
                }
                "xpan" - "ypan" {
                    PanCamera
                }
                "qx" - "qy" - "qz" - "qw" {
                    set q [list $_view(qw) $_view(qx) $_view(qy) $_view(qz)]
                    $_arcball quaternion $q
                    EventuallyRotate $q
                }
                "zoom" {
                    SendCmd "camera zoom $_view(zoom)"
                }
            }
        }
    }
}

itcl::body Rappture::VtkHeightmapViewer::GetVtkData { args } {
    set bytes ""
    foreach dataobj [get] {
        foreach comp [$dataobj components] {
            set tag $dataobj-$comp
            set contents [ConvertToVtkData $dataobj $comp]
            #set contents [$dataobj vtkdata $comp]
            append bytes "$contents\n\n"
        }
    }
    return [list .vtk $bytes]
}

itcl::body Rappture::VtkHeightmapViewer::GetImage { args } {
    if { [image width $_image(download)] > 0 && 
         [image height $_image(download)] > 0 } {
        set bytes [$_image(download) data -format "jpeg -quality 100"]
        set bytes [Rappture::encoding::decode -as b64 $bytes]
        return [list .jpg $bytes]
    }
    return ""
}

itcl::body Rappture::VtkHeightmapViewer::BuildDownloadPopup { popup command } {
    Rappture::Balloon $popup \
        -title "[Rappture::filexfer::label downloadWord] as..."
    set inner [$popup component inner]
    label $inner.summary -text "" -anchor w 
    radiobutton $inner.vtk_button -text "VTK data file" \
        -variable [itcl::scope _downloadPopup(format)] \
        -font "Helvetica 9 " \
        -value vtk  
    Rappture::Tooltip::for $inner.vtk_button "Save as VTK data file."
    radiobutton $inner.image_button -text "Image File" \
        -variable [itcl::scope _downloadPopup(format)] \
        -value image 
    Rappture::Tooltip::for $inner.image_button \
        "Save as digital image."

    button $inner.ok -text "Save" \
        -highlightthickness 0 -pady 2 -padx 3 \
        -command $command \
        -compound left \
        -image [Rappture::icon download]

    button $inner.cancel -text "Cancel" \
        -highlightthickness 0 -pady 2 -padx 3 \
        -command [list $popup deactivate] \
        -compound left \
        -image [Rappture::icon cancel]

    blt::table $inner \
        0,0 $inner.summary -cspan 2  \
        1,0 $inner.vtk_button -anchor w -cspan 2 -padx { 4 0 } \
        2,0 $inner.image_button -anchor w -cspan 2 -padx { 4 0 } \
        4,1 $inner.cancel -width .9i -fill y \
        4,0 $inner.ok -padx 2 -width .9i -fill y 
    blt::table configure $inner r3 -height 4
    blt::table configure $inner r4 -pady 4
    raise $inner.image_button
    $inner.vtk_button invoke
    return $inner
}

itcl::body Rappture::VtkHeightmapViewer::SetObjectStyle { dataobj comp } {
    # Parse style string.
    set tag $dataobj-$comp
    array set settings {
        -color BCGYR
        -edges 0
        -edgecolor black
        -lighting 1
        -linewidth 1.0
        -levels 10
        -visible 1
    }
    set style [$dataobj style $comp]
    array set settings $style
    # Save the initial component style.
    if { ![info exists _initialStyle($tag)] } {
	set _initialStyle($tag) [array get settings]
    }
    set _style($tag) [array get settings]
    if { $dataobj != $_first } {
        set settings(-opacity) 1
    }
    $itk_component(colormap) value $settings(-color)
    set _numContours $settings(-levels)
    set scale [GetHeightmapScale]
    SendCmd "heightmap add numcontours $settings(-levels) $scale $tag"
    set _comp2scale($tag) $_settings(heightmap-scale)
    SendCmd "heightmap edges $_settings(contour-edges) $tag"
    SendCmd "heightmap wireframe $_settings(contour-wireframe) $tag"
    SetColormap $dataobj $comp
}

itcl::body Rappture::VtkHeightmapViewer::IsValidObject { dataobj } {
    if {[catch {$dataobj isa Rappture::Field} valid] != 0 || !$valid} {
        return 0
    }
    return 1
}

# ----------------------------------------------------------------------
# USAGE: ReceiveLegend <colormap> <title> <min> <max> <size>
#
# Invoked automatically whenever the "legend" command comes in from
# the rendering server.  Indicates that binary image data with the
# specified <size> will follow.
# ----------------------------------------------------------------------
itcl::body Rappture::VtkHeightmapViewer::ReceiveLegend { colormap title min max size } {
    set _legendPending 0
    #puts stderr "ReceiveLegend colormap=$colormap title=$title range=$min,$max size=$size"
    set _title $title
    regsub {\(mag\)} $title "" _title 
    if { [IsConnected] } {
        set bytes [ReceiveBytes $size]
        if { ![info exists _image(legend)] } {
            set _image(legend) [image create photo]
        }
        $_image(legend) configure -data $bytes
        #puts stderr "read $size bytes for [image width $_image(legend)]x[image height $_image(legend)] legend>"
        if { [catch {DrawLegend $_title} errs] != 0 } {
	    global errorInfo
	    puts stderr "errs=$errs errorInfo=$errorInfo"
        }
    }
}

#
# DrawLegend --
#
#       Draws the legend in it's own canvas which resides to the right
#       of the contour plot area.
#
itcl::body Rappture::VtkHeightmapViewer::DrawLegend { name } {
    set c $itk_component(view)
    set w [winfo width $c]
    set h [winfo height $c]
    set font "Arial 8"
    set lineht [font metrics $font -linespace]
    
    if { $name == "component" } {
	set title ""
    } else {
	if { [info exists _fields($name)] } {
	    foreach { title units } $_fields($name) break
	    if { $units != "" } {
		set title [format "%s (%s)" $title $units]
	    }
	} else {
	    set title $name
	}
    }
    set x [expr $w - 2]
    if { !$_settings(legend-visible) } {
	$c delete legend
	return
    } 
    if { [$c find withtag "legend"] == "" } {
	set y 2 
	# If there's a legend title, create a text item for the title.
	if { $title != "" } {
	    $c create text $x $y \
		-anchor ne \
		-fill $itk_option(-plotforeground) -tags "title legend" \
		-font $font 
	    incr y $lineht
	}
	$c create text $x $y \
	    -anchor ne \
	    -fill $itk_option(-plotforeground) -tags "zmax legend" \
	    -font $font
	incr y $lineht
	$c create image $x $y \
	    -anchor ne \
	    -image $_image(legend) -tags "colormap legend"
	$c create text $x [expr {$h-2}] \
	    -anchor se \
	    -fill $itk_option(-plotforeground) -tags "zmin legend" \
	    -font $font
	#$c bind colormap <Enter> [itcl::code $this EnterLegend %x %y]
	$c bind colormap <Leave> [itcl::code $this LeaveLegend]
	$c bind colormap <Motion> [itcl::code $this MotionLegend %x %y]
    }
    $c delete isoline
    set x2 $x
    set iw [image width $_image(legend)]
    set ih [image height $_image(legend)]
    set x1 [expr $x2 - ($iw*12)/10]
    set color [$itk_component(isolinecolor) value]
    # Draw the isolines on the legend.
    if { $color != "none" } {
	set pixels [blt::vector create \#auto]
	set values [blt::vector create \#auto]
	set range [image height $_image(legend)]
	# Order of pixels is max to min (max is at top of legend).
	$pixels seq $ih 0 $_numContours
	set offset [expr 2 + $lineht]
	# If there's a legend title, increase the offset by the line height.
	if { $title != "" } {
	    incr offset $lineht
	}
	$pixels expr {round($pixels + $offset)}
	# Order of values is min to max.
	$values seq $_limits(zmin) $_limits(zmax) $_numContours
	set tags "isoline legend"
	foreach pos [$pixels range 0 end] value [$values range end 0] {
	    set y1 [expr int($pos)]
	    set id [$c create line $x1 $y1 $x2 $y1 -fill $color -tags $tags]
	    $c bind $id <Enter> [itcl::code $this EnterIsoline %x %y $value]
	    $c bind $id <Leave> [itcl::code $this LeaveIsoline]
	}
	blt::vector destroy $pixels $values
    }

    $c bind title <ButtonPress> [itcl::code $this Combo post]
    $c bind title <Enter> [itcl::code $this Combo activate]
    $c bind title <Leave> [itcl::code $this Combo deactivate]
    # Reset the item coordinates according the current size of the plot.
    $c itemconfigure title -text $title
    if { [info exists _limits(zmin)] && $_limits(zmin) != "" } {
	$c itemconfigure zmin -text [format %g $_limits(zmin)]
    }
    if { [info exists _limits(zmax)] && $_limits(zmax) != "" } {
	$c itemconfigure zmax -text [format %g $_limits(zmax)]
    }
    set y 2
    # If there's a legend title, move the title to the correct position
    if { $title != "" } {
	$c coords title $x $y
	incr y $lineht
    }
    $c coords zmax $x $y
    incr y $lineht
    $c coords colormap $x $y
    $c coords zmin $x [expr {$h - 2}]
}

#
# EnterIsoline --
#
itcl::body Rappture::VtkHeightmapViewer::EnterIsoline { x y value } {
    SetIsolineTip $x $y $value
}

#
# LeaveIsoline --
#
itcl::body Rappture::VtkHeightmapViewer::LeaveIsoline { } {
    Rappture::Tooltip::tooltip cancel
    .rappturetooltip configure -icon ""
}

#
# SetIsolineTip --
#
itcl::body Rappture::VtkHeightmapViewer::SetIsolineTip { x y value } {
    set c $itk_component(view)
    set w [winfo width $c]
    set h [winfo height $c]
    set font "Arial 8"
    set lineht [font metrics $font -linespace]
    
    if { [info exists _fields($_title)] } {
        foreach { title units } $_fields($_title) break
        if { $units != "" } {
            set title [format "%s (%s)" $title $units]
        }
    } else {
        set title $_title
    }
    set imgHeight [image height $_image(legend)]
    set coords [$c coords colormap]
    set imgX [expr $w - [image width $_image(legend)] - 2]
    set imgY [expr $y - 2 * ($lineht + 2)]

    .rappturetooltip configure -icon ""

    # Compute the value of the point
    set tipx [expr $x + 15] 
    set tipy [expr $y - 5]
    Rappture::Tooltip::text $c "Isoline $value"
    Rappture::Tooltip::tooltip show $c +$tipx,+$tipy    
}


#
# EnterLegend --
#
itcl::body Rappture::VtkHeightmapViewer::EnterLegend { x y } {
    SetLegendTip $x $y
}

#
# MotionLegend --
#
itcl::body Rappture::VtkHeightmapViewer::MotionLegend { x y } {
    Rappture::Tooltip::tooltip cancel
    set c $itk_component(view)
    SetLegendTip $x $y
}

#
# LeaveLegend --
#
itcl::body Rappture::VtkHeightmapViewer::LeaveLegend { } {
    Rappture::Tooltip::tooltip cancel
    .rappturetooltip configure -icon ""
}

#
# SetLegendTip --
#
itcl::body Rappture::VtkHeightmapViewer::SetLegendTip { x y } {
    set c $itk_component(view)
    set w [winfo width $c]
    set h [winfo height $c]
    set font "Arial 8"
    set lineht [font metrics $font -linespace]
    
    set imgHeight [image height $_image(legend)]
    set coords [$c coords colormap]
    set imgX [expr $w - [image width $_image(legend)] - 2]
    set imgY [expr $y - 2 * ($lineht + 2)]

    if { [info exists _fields($_title)] } {
        foreach { title units } $_fields($_title) break
        if { $units != "" } {
            set title [format "%s (%s)" $title $units]
        }
    } else {
        set title $_title
    }
    # Make a swatch of the selected color
    if { [catch { $_image(legend) get 10 $imgY } pixel] != 0 } {
        #puts stderr "out of range: $imgY"
        return
    }
    if { ![info exists _image(swatch)] } {
        set _image(swatch) [image create photo -width 24 -height 24]
    }
    set color [eval format "\#%02x%02x%02x" $pixel]
    $_image(swatch) put black  -to 0 0 23 23 
    $_image(swatch) put $color -to 1 1 22 22 
    .rappturetooltip configure -icon $_image(swatch)

    # Compute the value of the point
    if { [info exists _limits(zmax)] && [info exists _limits(zmin)] } {
        set t [expr 1.0 - (double($imgY) / double($imgHeight-1))]
        set value [expr $t * ($_limits(zmax) - $_limits(zmin)) + $_limits(zmin)]
    } else {
        set value 0.0
    }
    set tipx [expr $x + 15] 
    set tipy [expr $y - 5]
    Rappture::Tooltip::text $c "$title $value"
    Rappture::Tooltip::tooltip show $c +$tipx,+$tipy    
}

# ----------------------------------------------------------------------
# USAGE: _dropdown post
# USAGE: _dropdown unpost
# USAGE: _dropdown select
#
# Used internally to handle the dropdown list for this combobox.  The
# post/unpost options are invoked when the list is posted or unposted
# to manage the relief of the controlling button.  The select option
# is invoked whenever there is a selection from the list, to assign
# the value back to the gauge.
# ----------------------------------------------------------------------
itcl::body Rappture::VtkHeightmapViewer::Combo {option} {
    set c $itk_component(view) 
    switch -- $option {
        post {
            foreach { x1 y1 x2 y2 } [$c bbox title] break
            set x1 [expr [winfo width $itk_component(view)] - [winfo reqwidth $itk_component(fieldmenu)]]
            set x [expr $x1 + [winfo rootx $itk_component(view)]]
            set y [expr $y2 + [winfo rooty $itk_component(view)]]
            tk_popup $itk_component(fieldmenu) $x $y
        }
        activate {
            $c itemconfigure title -fill red
        }
        deactivate {
            $c itemconfigure title -fill $itk_option(-plotforeground) 
        }
        invoke {
            $itk_component(field) value $_currentField
            AdjustSetting contour-field
        }
        default {
            error "bad option \"$option\": should be post, unpost, select"
        }
    }
}

itcl::body Rappture::VtkHeightmapViewer::GetHeightmapScale {} {
    if {  $_settings(heightmap-mode) } {
	set val $_settings(heightmap-scale)
	set sval [expr { $val >= 50 ? double($val)/50.0 : 1.0/(2.0-(double($val)/50.0)) }]
	return $sval
    }
    set sval 0 
}

itcl::body Rappture::VtkHeightmapViewer::GetColormap { color tag } {
    array set style {
	-color BCGYR
	-levels 10
	-opacity 1.0
    }
    if { ![info exists _style($tag)] } {
	set _style($tag) [array get style]
    }
    array set style $_style($tag)
    set style(-color) $color
    set _style($tag) [array get style]
    array set style $_style($tag)
    set name "$style(-color):$style(-levels):$style(-opacity)"
    if { ![info exists _colormaps($name)] } {
	BuildColormap $name [array get style]
	set _colormaps($name) 1
    }
    return $name
}
